#
# Copyright (C) 2005-2008 by Pieter Palmers
#               2007-2008 by Arnold Krille
#
# This file is part of FFADO
# FFADO = Free FireWire (pro-)audio drivers for Linux
#
# FFADO is based upon FreeBoB.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import ffado.config
from ffado.configuration import *

import dbus
try:
    # First try the PyQt4 module name
    from dbus.mainloop.qt import DBusQtMainLoop
except ImportError:
    from dbus.mainloop.pyqt5 import DBusQtMainLoop

DBusQtMainLoop(set_as_default=True)

import logging
log = logging.getLogger('dbus')

class ControlInterface:
    def __init__(self, servername, basepath):
        self.basepath=basepath
        self.servername=servername
        if ffado.config.bypassdbus:
            self.devices = DeviceList(ffado.config.SYSTEM_CONFIG_FILE)
            self.devices.updateFromFile(ffado.config.USER_CONFIG_FILE)
        else:
            self.bus=dbus.SessionBus()

    def setContignuous(self, subpath, v, idx=None):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would set Continuous %s on server %s" % (path, self.servername))
            return
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Continuous')
            if idx == None:
                dev_cont.setValue(v)
            else:
                dev_cont.setValueIdx(idx,v)
        except:
            log.error("Failed to set Continuous %s on server %s" % (path, self.servername))

    def getContignuous(self, subpath, idx=None):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would get Continuous %s on server %s" % (path, self.servername))
            return 0
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Continuous')
            if idx == None:
                return int(dev_cont.getValue())
            else:
                return int(dev_cont.getValueIdx(idx))
        except:
            log.error("Failed to get Continuous %s on server %s" % (path, self.servername))
            return 0

    def setDiscrete(self, subpath, v, idx=None):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would set Discrete %s on server %s" % (path, self.servername))
            return
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Discrete')
            if idx == None:
                dev_cont.setValue(v)
            else:
                dev_cont.setValueIdx(v, idx)
        except:
            log.error("Failed to set Discrete %s on server %s" % (path, self.servername))

    def getDiscrete(self, subpath, idx=None):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
             log.info("bypassdbus set, would get Discrete %s on server %s" % (path, self.servername))
             return 0
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Discrete')
            if idx == None:
                return int(dev_cont.getValue())
            else:
                return int(dev_cont.getValueIdx(idx))
        except:
            log.error("Failed to get Discrete %s on server %s" % (path, self.servername))
            return 0

    def setText(self, subpath, v):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("DEBUG_BYPASSDBUS set, would set Text %s on server %s" % (path, self.servername))
            return
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Text')
            dev_cont.setValue(v)
        except:
            log.error("Failed to set Text %s on server %s" % (path, self.servername))

    def getText(self, subpath):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would get get Text %s on server %s" % (path, self.servername))
            return ""
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Text')
            return dev_cont.getValue()
        except:
            log.error("Failed to get Text %s on server %s" % (path, self.servername))
            return 0

    def setMatrixMixerValue(self, subpath, row, col, v):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would set MatrixMixer %s on server %s" % (path, self.servername))
            return
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.MatrixMixer')
            dev_cont.setValue(row, col, v)
        except:
            log.error("Failed to set MatrixMixer %s on server %s" % (path, self.servername))

    def getMatrixMixerValue(self, subpath, row, col):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would get MatrixMixer %s on server %s" % (path, self.servername))
            return 0
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.MatrixMixer')
            return int(dev_cont.getValue(row, col))
        except:
            log.error("Failed to get MatrixMixer %s on server %s" % (path, self.servername))
            return 0

    def enumSelect(self, subpath, v):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would select %s on server %s" % (path, self.servername))
            return
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Enum')
            dev_cont.select(v)
        except:
            log.error("Failed to select %s on server %s" % (path, self.servername))

    def enumSelected(self, subpath):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would get selected enum %s on server %s" % (path, self.servername))
            return 0
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Enum')
            return dev_cont.selected()
        except:
            log.error("Failed to get selected enum %s on server %s" % (path, self.servername))
            return 0

    def enumGetLabel(self, subpath, v):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would get enum label %s on server %s" % (path, self.servername))
            return 0
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Enum')
            return dev_cont.getEnumLabel(v)
        except:
            log.error("Failed to get enum label %s on server %s" % (path, self.servername))
            return 0

    def enumCount(self, subpath):
        path = self.basepath + subpath
        if ffado.config.bypassdbus:
            log.info("bypassdbus set, would get enum count %s on server %s" % (path, self.servername))
            return 0
        try:
            dev = self.bus.get_object(self.servername, path)
            dev_cont = dbus.Interface(dev, dbus_interface='org.ffado.Control.Element.Enum')
            return dev_cont.count()
        except:
            log.error("Failed to get enum count %s on server %s" % (path, self.servername))
            return 0

class DeviceManagerInterface:
    """ Implementation of the singleton """
    def __init__(self, servername, basepath):
        self.basepath=basepath + '/DeviceManager'
        self.servername=servername
        if ffado.config.bypassdbus:
            self.devices = DeviceList(ffado.config.SYSTEM_CONFIG_FILE)
            self.devices.updateFromFile(ffado.config.USER_CONFIG_FILE)
        else:
            self.bus=dbus.SessionBus()
            self.dev = self.bus.get_object(self.servername, self.basepath)
            self.iface = dbus.Interface(self.dev, dbus_interface='org.ffado.Control.Element.Container')

        self.updateSignalHandlers = []
        self.updateSignalHandlerArgs = {}
        self.preUpdateSignalHandlers = []
        self.preUpdateSignalHandlerArgs = {}
        self.postUpdateSignalHandlers = []
        self.postUpdateSignalHandlerArgs = {}
        self.destroyedSignalHandlers = []
        self.destroyedSignalHandlerArgs = {}

        if ffado.config.bypassdbus:
            return

        # signal reception does not work yet since we need a mainloop for that
        # and qt3 doesn't provide one for python/dbus
        try:
            log.debug("connecting to: Updated on %s (server: %s)" % (self.basepath, self.servername))
            self.dev.connect_to_signal("Updated", self.updateSignal, \
                                    dbus_interface="org.ffado.Control.Element.Container")
            self.dev.connect_to_signal("PreUpdate", self.preUpdateSignal, \
                                    dbus_interface="org.ffado.Control.Element.Container")
            self.dev.connect_to_signal("PostUpdate", self.postUpdateSignal, \
                                    dbus_interface="org.ffado.Control.Element.Container")
            self.dev.connect_to_signal("Destroyed", self.destroyedSignal, \
                                    dbus_interface="org.ffado.Control.Element.Container")

        except dbus.DBusException:
            traceback.print_exc()

    def registerPreUpdateCallback(self, callback, arg=None):
        if not callback in self.preUpdateSignalHandlers:
            self.preUpdateSignalHandlers.append(callback)
        # always update the argument
        self.preUpdateSignalHandlerArgs[callback] = arg

    def registerPostUpdateCallback(self, callback, arg=None):
        if not callback in self.postUpdateSignalHandlers:
            self.postUpdateSignalHandlers.append(callback)
        # always update the argument
        self.postUpdateSignalHandlerArgs[callback] = arg

    def registerUpdateCallback(self, callback, arg=None):
        if not callback in self.updateSignalHandlers:
            self.updateSignalHandlers.append(callback)
        # always update the argument
        self.updateSignalHandlerArgs[callback] = arg

    def registerDestroyedCallback(self, callback, arg=None):
        if not callback in self.destroyedSignalHandlers:
            self.destroyedSignalHandlers.append(callback)
        # always update the argument
        self.destroyedSignalHandlerArgs[callback] = arg

    def updateSignal(self):
        log.debug("Received update signal")
        for handler in self.updateSignalHandlers:
            arg = self.updateSignalHandlerArgs[handler]
            try:
                if arg:
                    handler(arg)
                else:
                    handler()
            except:
                log.error("Failed to execute handler %s" % handler)

    def preUpdateSignal(self):
        log.debug("Received pre-update signal")
        for handler in self.preUpdateSignalHandlers:
            arg = self.preUpdateSignalHandlerArgs[handler]
            try:
                if arg:
                    handler(arg)
                else:
                    handler()
            except:
                log.error("Failed to execute handler %s" % handler)

    def postUpdateSignal(self):
        log.debug("Received post-update signal")
        for handler in self.postUpdateSignalHandlers:
            arg = self.postUpdateSignalHandlerArgs[handler]
            try:
                if arg:
                    handler(arg)
                else:
                    handler()
            except:
                log.error("Failed to execute handler %s" % handler)

    def destroyedSignal(self):
        log.debug("Received destroyed signal")
        for handler in self.destroyedSignalHandlers:
            arg = self.destroyedSignalHandlerArgs[handler]
            try:
                if arg:
                    handler(arg)
                else:
                    handler()
            except:
                log.error("Failed to execute handler %s" % handler)

    def getNbDevices(self):
        if ffado.config.bypassdbus:
            return len(self.devices.devices)
        return self.iface.getNbElements()
    def getDeviceName(self, idx):
        if ffado.config.bypassdbus:
            return str(idx)
        return self.iface.getElementName(idx)

class ConfigRomInterface:
    def __init__(self, servername, devicepath):
        self.basepath=devicepath + '/ConfigRom'
        self.servername=servername
        if ffado.config.bypassdbus:
            self.devices = DeviceList(ffado.config.SYSTEM_CONFIG_FILE)
            self.devices.updateFromFile(ffado.config.USER_CONFIG_FILE)
            self.idx = int(devicepath)
        else:
            self.bus=dbus.SessionBus()
            self.dev = self.bus.get_object(self.servername, self.basepath)
            self.iface = dbus.Interface(self.dev, dbus_interface='org.ffado.Control.Element.ConfigRomX')
    def getGUID(self):
        if ffado.config.bypassdbus:
            return str(self.idx)
        return self.iface.getGUID()
    def getVendorName(self):
        if ffado.config.bypassdbus:
            return self.devices.devices[self.idx]['vendorname']
        return self.iface.getVendorName()
    def getModelName(self):
        if ffado.config.bypassdbus:
            return self.devices.devices[self.idx]['modelname']
        return self.iface.getModelName()
    def getVendorId(self):
        if ffado.config.bypassdbus:
            return int(self.devices.devices[self.idx]['vendorid'], 16)
        return self.iface.getVendorId()
    def getModelId(self):
        if ffado.config.bypassdbus:
            return int(self.devices.devices[self.idx]['modelid'], 16)
        return self.iface.getModelId()
    def getUnitVersion(self):
        if ffado.config.bypassdbus:
            return 0
        return self.iface.getUnitVersion()

class ClockSelectInterface:
    def __init__(self, servername, devicepath):
        self.basepath=devicepath + '/Generic/ClockSelect'
        self.servername=servername
        if ffado.config.bypassdbus:
            self.devices = DeviceList(ffado.config.SYSTEM_CONFIG_FILE)
            self.devices.updateFromFile(ffado.config.USER_CONFIG_FILE)
            self.idx = devicepath
        else:
            self.bus=dbus.SessionBus()
            self.dev = self.bus.get_object(self.servername, self.basepath)
            self.iface = dbus.Interface(self.dev, dbus_interface='org.ffado.Control.Element.AttributeEnum')
            self.iface_element = dbus.Interface(self.dev, dbus_interface='org.ffado.Control.Element.Element')
    def count(self):
        if ffado.config.bypassdbus:
            return 1
        return self.iface.count()
    def select(self, idx):
        if ffado.config.bypassdbus:
            return 1
        return self.iface.select(idx)
    def selected(self):
        if ffado.config.bypassdbus:
            return True
        return self.iface.selected()
    def getEnumLabel(self, idx):
        if ffado.config.bypassdbus:
            return 'enumlabel ' + str(idx)
        return self.iface.getEnumLabel(idx)
    def attributeCount(self):
        if ffado.config.bypassdbus:
            return 1
        return self.iface.attributeCount()
    def getAttributeValue(self, idx):
        if ffado.config.bypassdbus:
            return 1
        return self.iface.getAttributeValue(idx)
    def getAttributeName(self, idx):
        if ffado.config.bypassdbus:
            return 'attrib ' + str(idx)
        return self.iface.getAttributeName(idx)
    def canChangeValue(self):
        if ffado.config.bypassdbus:
            return 1
        return self.iface_element.canChangeValue()

class EnumInterface:
    def __init__(self, servername, basepath):
        self.basepath = basepath
        self.servername = servername
        if ffado.config.bypassdbus:
            self.devices = DeviceList(ffado.config.SYSTEM_CONFIG_FILE)
            self.devices.updateFromFile(ffado.config.USER_CONFIG_FILE)
        else:
            self.bus = dbus.SessionBus()
            self.dev = self.bus.get_object(self.servername, self.basepath)
            self.iface = dbus.Interface(self.dev, dbus_interface='org.ffado.Control.Element.Enum')
            self.iface_element = dbus.Interface(self.dev, dbus_interface='org.ffado.Control.Element.Element')
    def count(self):
        if ffado.config.bypassdbus:
            return 1
        return self.iface.count()
    def select(self, idx):
        if ffado.config.bypassdbus:
            return 1
        return self.iface.select(idx)
    def selected(self):
        if ffado.config.bypassdbus:
            return True
        return self.iface.selected()
    def getEnumLabel(self, idx):
        if ffado.config.bypassdbus:
            # Can't include text here since some code uses int() to extract
            # a value from the enum label
            return '0'
        return self.iface.getEnumLabel(idx)
    def canChangeValue(self):
        if ffado.config.bypassdbus:
            return True
        return self.iface_element.canChangeValue()
    def devConfigChanged(self, idx):
        if ffado.config.bypassdbus:
            return True
        return self.iface.devConfigChanged(idx)

class SamplerateSelectInterface(EnumInterface):
    def __init__(self, servername, devicepath):
        EnumInterface.__init__(self, servername, devicepath + '/Generic/SamplerateSelect')

class StreamingStatusInterface(EnumInterface):
    def __init__(self, servername, devicepath):
        EnumInterface.__init__(self, servername, devicepath + '/Generic/StreamingStatus')

class TextInterface:
    def __init__(self, servername, basepath):
        self.basepath=basepath
        self.servername=servername
        if ffado.config.bypassdbus:
            self.devices = DeviceList(ffado.config.SYSTEM_CONFIG_FILE)
            self.devices.updateFromFile(ffado.config.USER_CONFIG_FILE)
        else:
            self.bus=dbus.SessionBus()
            self.dev = self.bus.get_object( self.servername, self.basepath )
            self.iface = dbus.Interface( self.dev, dbus_interface="org.ffado.Control.Element.Text" )
            self.iface_element = dbus.Interface(self.dev, dbus_interface='org.ffado.Control.Element.Element')
    def text(self):
        if ffado.config.bypassdbus:
            return "text"
        return self.iface.getValue()
    def setText(self,text):
        if ffado.config.bypassdbus:
            return
        self.iface.setValue(text)
    def canChangeValue(self):
        if ffado.config.bypassdbus:
            return True
        return self.iface_element.canChangeValue()

# vim: et



# coding=utf8
#
# Copyright (C) 2009 by Arnold Krille
#               2013 by Philippe Carriere
#
# This file is part of FFADO
# FFADO = Free FireWire (pro-)audio drivers for Linux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# from PyQt4 import QtGui, QtCore, Qt
# from PyQt4.QtCore import pyqtSignal
# from PyQt4.QtGui import QColor, QAbstractSlider, QDoubleSpinBox, QWidgetAction
# from PyQt4.QtGui import QAction, QPainter, QWidget, QGridLayout, QLabel
# from PyQt4.QtGui import QLayout, QSlider, QLineEdit, QPalette
# from PyQt4.QtGui import QVBoxLayout, QHBoxLayout, QTabWidget, QToolBar
# from PyQt4.QtGui import QComboBox, QScrollArea, QPushButton, QSizePolicy
from ffado.import_pyqt import *

import dbus, math, decimal

import ffado.config

import logging
log = logging.getLogger("matrixmixer")

def toDBvalue(value):
    n = int(value)
    c2p14 = 16384.0
    if n > 164:
        return round(20.0*math.log10(float(n)/c2p14), 2)
    else:
        return -40.0

def fromDBvalue(value):
    v = float(value)
    c2p14 = 16384.0
    if (v > -40):
        return int(round(math.pow(10.0, (value/20.0))*c2p14, 0))
    else:
        return 0

# v, vl, vr in linear scale
# b range in [-1:1]
def getVolumeLeft(v, b):
    return int(round(0.5*v*(1.0-b),0))
def getVolumeRight(v, b):
    return v-int(round(0.5*v*(1.0-b),0))
def getStereoVolume(vl, vr):
    return int(round(vl+vr,0))
def getStereoBalance(vl, vr):
    if ((vl+vr) == 0):
        return 0
    else:
        return round(float(vr-vl)/float(vr+vl),2)

class ColorForNumber:
    def __init__(self):
        self.colors = dict()

    def addColor(self, n, color):
        self.colors[n] = color

    def getColor(self, n):
        #print( "ColorForNumber.getColor( %g )" % (n) )
        keys = sorted(self.colors.keys())
        low = keys[-1]
        high = keys[-1]
        for i in range(len(keys)-1):
            if keys[i] <= n and keys[i+1] > n:
                low = keys[i]
                high = keys[i+1]
        #print( "%g is between %g and %g" % (n, low, high) )
        f = 0
        if high != low:
            f = (n-low) / (high-low)
        lc = self.colors[low]
        hc = self.colors[high]
        return QColor(
                int((1-f)*lc.red()   + f*hc.red()),
                int((1-f)*lc.green() + f*hc.green()),
                int((1-f)*lc.blue()  + f*hc.blue()) )

class BckgrdColorForNumber(ColorForNumber):
    def __init__(self):
        ColorForNumber.__init__(self)
        self.addColor(             0.0, QColor(  0,   0,   0))
        self.addColor(             1.0, QColor(  0,   0, 128))
        self.addColor(   math.pow(2,6), QColor(  0, 255,   0))
        self.addColor(  math.pow(2,14), QColor(255, 255,   0))
        self.addColor(math.pow(2,16)-1, QColor(255,   0,   0))

    def getFrgdColor(self, color):
        if color.valueF() < 0.6:
            return QColor(255, 255, 255)
        else:
            return QColor(0, 0, 0)
    
class MixerNode(QAbstractSlider):
    nodeValueChanged = pyqtSignal(tuple)
    def __init__(self, input, output, value, max, muted, inverted, parent, matrix_obj):
        QAbstractSlider.__init__(self, parent)
        #log.debug("MixerNode.__init__( %i, %i, %i, %i, %s )" % (input, output, value, max, str(parent)) )

        # Store a direct link back to the underlying matrix object so the mute
        # and invert interfaces can be easily found.  By the time the matrix 
        # has been set into the full widget hierarchy, its parent is unlikely
        # to still be the top-level matrix object.
        self.matrix_obj = matrix_obj;

        self.pos = QtCore.QPointF(0, 0)
        self.input = input
        self.output = output
        self.setOrientation(Qt.Vertical)
        if max == -1:
            max = pow(2, 16)-1
        self.setRange(0, max)
        self.setValue(int(value))
        self.valueChanged.connect(self.internalValueChanged)

        self.setSmall(False)

        self.bgcolors = BckgrdColorForNumber()

        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.mapper = QtCore.QSignalMapper(self)
        self.mapper.mapped['QString'].connect(self.directValues)

        self.spinbox = QDoubleSpinBox(self)
        self.spinbox.setRange(-40, 12)
        self.spinbox.setSuffix(" dB")
        if value != 0:
            self.spinbox.setValue(toDBvalue(value))            

        self.spinbox.valueChanged.connect(self.directValues)
        action = QWidgetAction(self)
        action.setDefaultWidget(self.spinbox)
        self.addAction(action)

        for text in ["3 dB", "0 dB", "-3 dB", "-20 dB", "-inf dB"]:
            action = QAction(text, self)
            action.triggered.connect(self.mapper.map)
            self.mapper.setMapping(action, text)
            self.addAction(action)

        # Only show the mute menu item if a value has been supplied
        self.mute_action = None
        if (muted != None):
            action = QAction(text, self)
            action.setSeparator(True)
            self.addAction(action)
            self.mute_action = QAction("Mute", self)
            self.mute_action.setCheckable(True)
            self.mute_action.setChecked(bool(muted))
            self.mute_action.triggered.connect(self.mapper.map)
            self.mapper.setMapping(self.mute_action, "Mute")
            self.addAction(self.mute_action)

        # Similarly, only show a phase inversion menu item if in use
        self.inv_action = None
        if (inverted != None):
            if (muted == None):
                action = QAction(text, self)
                action.setSeparator(True)
                self.addAction(action)
            self.inv_action = QAction("Invert", self)
            self.inv_action.setCheckable(True)
            self.inv_action.setChecked(bool(inverted))
            self.inv_action.triggered.connect(self.mapper.map)
            self.mapper.setMapping(self.inv_action, "Invert")
            self.addAction(self.inv_action)

    def directValues(self,text):
        #log.debug("MixerNode.directValues( '%s' )" % text)
        if text == "Mute":
            #log.debug("Mute %d" % self.mute_action.isChecked())
            self.update()
            self.matrix_obj.mutes_interface.setValue(self.output, self.input, self.mute_action.isChecked())
        elif text == "Invert":
            #log.debug("Invert %d" % self.inv_action.isChecked())
            self.update()
            self.matrix_obj.inverts_interface.setValue(self.output, self.input, self.inv_action.isChecked())
        else:
            text = str(text).split(" ")[0].replace(",",".")
            n = fromDBvalue(float(text))
            #log.debug("  linear value: %g" % n)
            self.setValue(n)

    def mousePressEvent(self, ev):
        if ev.buttons() & Qt.LeftButton:
            self.pos = ev.posF() if ffado_pyqt_version == 4 else ev.localPos()
            self.tmpvalue = self.value()
            ev.accept()
            #log.debug("MixerNode.mousePressEvent() %s" % str(self.pos))

    def mouseMoveEvent(self, ev):
        if hasattr(self, "tmpvalue") and self.pos is not QtCore.QPointF(0, 0):
            newpos = ev.posF() if ffado_pyqt_version == 4 else ev.localPos()
            change = newpos.y() - self.pos.y()
            #log.debug("MixerNode.mouseReleaseEvent() change %s" % (str(change)))
            self.setValue(
                int(self.tmpvalue - math.copysign(pow(abs(change), 2), change))
            )
            ev.accept()

    def mouseReleaseEvent(self, ev):
        if hasattr(self, "tmpvalue") and self.pos is not QtCore.QPointF(0, 0):
            newpos = ev.posF() if ffado_pyqt_version == 4 else ev.localPos()
            change = newpos.y() - self.pos.y()
            #log.debug("MixerNode.mouseReleaseEvent() change %s" % (str(change)))
            self.setValue(
                int(self.tmpvalue - math.copysign(pow(abs(change), 2), change))
            )
            self.pos = QtCore.QPointF(0, 0)
            del self.tmpvalue
            ev.accept()

    # Wheel event is mainly for scrolling inside the mixer window
    #   Additionnaly press Control key for wheel controling the values
    def wheelEvent (self, ev):
        if (ev.modifiers() & Qt.ControlModifier):
            tmpvalue = self.value()
            change = ev.delta()/8
            self.setValue(
                int(tmpvalue + math.copysign(pow(abs(change), 2), change))
            )
            ev.accept()
        else:
            ev.ignore()

    def paintEvent(self, ev):
        p = QPainter(self)
        rect = self.rect()
        v = self.value()
        if (self.mute_action!=None and self.mute_action.isChecked()):
            color = QColor(64, 64, 64)
        else:
            color = self.bgcolors.getColor(v)
        p.fillRect(rect, color)

        if self.small:
            return

        p.setPen(self.bgcolors.getFrgdColor(color))

        lv=decimal.Decimal('-Infinity')
        if v != 0:
            lv = toDBvalue(v)
            #log.debug("new value is %g dB" % lv)
        text = "%.2g dB" % lv
        if v == 0:
            symb_inf = u"\u221E"
            text = "-" + symb_inf + " dB"
        if ffado_python3 or ffado_pyqt_version == 5:
            # Python3 uses native python UTF strings rather than QString.
            # This therefore appears to be the correct way to display this
            # UTF8 string, but testing may prove otherwise.
            p.drawText(rect, Qt.AlignCenter, text)
        else:
            p.drawText(rect, Qt.AlignCenter, QString.fromUtf8(text))
        if (self.inv_action!=None and self.inv_action.isChecked()):
            if ffado_python3 or ffado_pyqt_version == 5:
                # Refer to the comment about about Python UTF8 strings.
                p.drawText(rect, Qt.AlignLeft|Qt.AlignTop, " ϕ")
            else:
                p.drawText(rect, Qt.AlignLeft|Qt.AlignTop, QString.fromUtf8(" ϕ"))

    def internalValueChanged(self, value):
        #log.debug("MixerNode.internalValueChanged( %i )" % value)
        if value != 0:
            dB = toDBvalue(value)
            if self.spinbox.value() is not dB:
                self.spinbox.setValue(dB)
        self.nodeValueChanged.emit((self.input, self.output, value))
        self.update()

    def setSmall(self, small):
        self.small = small
        if small:
            self.setMinimumSize(10, 10)
        else:
            fontmetrics = self.fontMetrics()
            self.setMinimumSize(fontmetrics.boundingRect("-0.0 dB").size()*1.1)
        self.update()

class MixerChannel(QWidget):
    hide = pyqtSignal(int, bool, name='hide')
    def __init__(self, number, parent=None, name="", smallFont=False):
        QWidget.__init__(self, parent)
        layout = QGridLayout(self)
        self.number = number
        self.name = name
        self.lbl = QLabel(self)
        self.lbl.setAlignment(Qt.AlignCenter)
        if (smallFont):
            font = self.lbl.font()
            font.setPointSize(int(font.pointSize()/1.5 + 0.5))
            self.lbl.setFont(font)
        layout.addWidget(self.lbl, 0, 0, 1, 2)
        self.hideChannel(False)

        self.setContextMenuPolicy(Qt.ActionsContextMenu)

        action = QAction("Make this channel small", self)
        action.setCheckable(True)
        action.triggered.connect(self.hideChannel)
        self.addAction(action)

    def hideChannel(self, hide):
        if hide:
            self.lbl.setText("%i" % (self.number+1));
        else:
            self.lbl.setText(self.name)
        self.hide.emit(self.number, hide)
        self.update()

# Matrix view widget
class MatrixControlView(QWidget):
    valueChanged = pyqtSignal([tuple])
    def __init__(self, servername, basepath, parent=None, sliderMaxValue=-1, mutespath=None, invertspath=None, smallFont=False, shortname=False, shortcolname="Ch", shortrowname="Ch", transpose=False):
        QWidget.__init__(self, parent)

        if not ffado.config.bypassdbus:
            self.bus = dbus.SessionBus()
            self.dev = self.bus.get_object(servername, basepath)
            self.interface = dbus.Interface(self.dev, dbus_interface="org.ffado.Control.Element.MatrixMixer")

        self.transpose = transpose
        if (transpose):
            self.shortcolname = shortrowname
            self.shortrowname = shortcolname
            if ffado.config.bypassdbus:
                self.cols = 2
                self.rows = 2
            else:
                self.cols = self.interface.getRowCount()
                self.rows = self.interface.getColCount()
        else:
            self.shortcolname = shortcolname
            self.shortrowname = shortrowname
            if ffado.config.bypassdbus:
                self.cols = 2
                self.rows = 2
            else:
                self.cols = self.interface.getColCount()
                self.rows = self.interface.getRowCount()

        log.debug("Mixer has %i rows and %i columns" % (self.rows, self.cols))

        self.mutes_dev = None
        self.mutes_interface = None
        if not ffado.config.bypassdbus and (mutespath != None):
            self.mutes_dev = self.bus.get_object(servername, mutespath)
            self.mutes_interface = dbus.Interface(self.mutes_dev, dbus_interface="org.ffado.Control.Element.MatrixMixer")

        self.inverts_dev = None
        self.inverts_interface = None
        if not ffado.config.bypassdbus and (invertspath != None):
            self.inverts_dev = self.bus.get_object(servername, invertspath)
            self.inverts_interface = dbus.Interface(self.inverts_dev, dbus_interface="org.ffado.Control.Element.MatrixMixer")

        layout = QGridLayout(self)
        layout.setSizeConstraint(QLayout.SetNoConstraint);
        self.setLayout(layout)

        self.rowHeaders = []
        self.columnHeaders = []
        self.items = []
        self.shortname = shortname

        # Add row/column headers, but only if there's more than one 
        # row/column
        if (self.cols > 1):
            for i in range(self.cols):
                ch = MixerChannel(i, self, self.getColName(i, self.shortname), smallFont)
                ch.hide.connect(self.hideColumn)
                layout.addWidget(ch, 0, i+1)
                self.columnHeaders.append( ch )
            layout.setRowStretch(0, 0)
            layout.setRowStretch(1, 10)
        if (self.rows > 1):
            for i in range(self.rows):
                ch = MixerChannel(i, self, self.getRowName(i, self.shortname), smallFont)
                ch.hide.connect(self.hideRow)
                layout.addWidget(ch, i+1, 0)
                self.rowHeaders.append( ch )

        # Add node-widgets
        for i in range(self.rows):
            self.items.append([])
            for j in range(self.cols):
                if (transpose):
                    mute_value = None
                    if (self.mutes_interface != None):
                        mute_value = self.mutes_interface.getValue(j,i)
                    inv_value = None
                    if (self.inverts_interface != None):
                        inv_value = self.inverts_interface.getValue(j,i)
                    if ffado.config.bypassdbus:
                        val = 0
                    else:
                        val = self.interface.getValue(j,i)
                    node = MixerNode(i, j, val, sliderMaxValue, mute_value, inv_value, self, self)
                else:
                    mute_value = None
                    if (self.mutes_interface != None):
                        mute_value = self.mutes_interface.getValue(i,j)
                    inv_value = None
                    if (self.inverts_interface != None):
                        inv_value = self.inverts_interface.getValue(i,j)
                    if ffado.config.bypassdbus:
                        val = 0
                    else:
                        val = self.interface.getValue(i,j)
                    node = MixerNode(j, i, val, sliderMaxValue, mute_value, inv_value, self, self)
                if (smallFont):
                    font = node.font()
                    font.setPointSize(int(font.pointSize()/1.5 + 0.5))
                    node.setFont(font)
                self.nodeConnect(node)
                layout.addWidget(node, i+1, j+1)
                self.items[i].append(node)

        self.hiddenRows = []
        self.hiddenCols = []

    def nodeConnect(self, node):
        node.nodeValueChanged.connect(self.valueChangedFn)

    def nodeDisconnect(self, node):
        node.nodeValueChanged.disconnect(self.valueChangedFn)

    def checkVisibilities(self):
        for x in range(len(self.items)):
            for y in range(len(self.items[x])):
                self.items[x][y].setSmall(
                        (x in self.hiddenRows)
                        | (y in self.hiddenCols)
                        )

    def hideColumn(self, column, hide):
        if hide:
            self.hiddenCols.append(column)
        else:
            self.hiddenCols.remove(column)
        self.checkVisibilities()

    def hideRow(self, row, hide):
        if hide:
            self.hiddenRows.append(row)
        else:
            self.hiddenRows.remove(row)
        self.checkVisibilities()

    # Columns and rows
    def getColName(self, i, shortname):
        if ffado.config.bypassdbus:
            return 'col ' + str(i)
        if (self.transpose):
            name = self.interface.getRowName(i)
        else:
            name = self.interface.getColName(i)
        self.shortname = shortname
        if (shortname or (name == '')):
            number = " %d" % (i+1)
            name = self.shortcolname + number
        return name

    def getRowName(self, j, shortname):
        if ffado.config.bypassdbus:
            return 'row ' + str(j)
        if (self.transpose):
            name = self.interface.getColName(j)
        else:
            name = self.interface.getRowName(j)
        self.shortname = shortname
        if (shortname or (name == '')):
            number = " %d" % (j+1)
            name = self.shortrowname + number
        return name

    def valueChangedFn(self, n):
        #log.debug("MatrixNode.valueChangedFn( %s )" % str(n))
        if not ffado.config.bypassdbus:
            self.interface.setValue(n[1], n[0], n[2])
        self.valueChanged.emit(n)
        
    # Update when routing is modified
    def updateRouting(self):
        if (self.cols > 1):
            for i in range(self.cols):
                last_name = self.columnHeaders[i].lbl.text()
                col_name = self.getColName(i, self.shortname)
                if last_name != col_name:
                    #log.debug("MatrixControlView.updateRouting( %s )" % str(col_name))
                    self.columnHeaders[i].name = col_name
                    self.columnHeaders[i].lbl.setText(col_name)
      
        if (self.rows > 1):
            for j in range(self.rows):
                last_name = self.rowHeaders[j].lbl.text()
                row_name = self.getRowName(j, self.shortname)
                if last_name != row_name:
                    #log.debug("MatrixControlView.updateRouting( %s )" % str(row_name))
                    self.rowHeaders[j].name = row_name
                    self.rowHeaders[j].lbl.setText(row_name)

    def updateValues(self, n):
        nbitems = len(n) // 3
        for i in range(nbitems):
            n_0 = n[3*i]    
            n_1 = n[3*i+1]   
            n_2 = n[3*i+2] 
            self.nodeDisconnect(self.items[n_0][n_1])
            self.items[n_0][n_1].setValue(n_2)
            self.nodeConnect(self.items[n_0][n_1])

    def refreshValues(self):
        if ffado.config.bypassdbus:
            return
        for x in range(len(self.items)):
            for y in range(len(self.items[x])):
                val = self.interface.getValue(x,y)
                if (self.transpose):
                    self.items[y][x].setValue(int(val))
                    self.items[y][x].internalValueChanged(val)
                else:
                    self.items[x][y].setValue(int(val))
                    self.items[x][y].internalValueChanged(val)

    def saveSettings(self, indent):
        matrixSaveString = []
        matrixSaveString.append('%s  <row_number>\n' % indent)
        matrixSaveString.append('%s    %d\n' % (indent, self.rows))        
        matrixSaveString.append('%s  </row_number>\n' % indent)
        matrixSaveString.append('%s  <col_number>\n' % indent)
        matrixSaveString.append('%s    %d\n' % (indent, self.cols))        
        matrixSaveString.append('%s  </col_number>\n' % indent)
        matrixSaveString.append('%s  <coefficients>\n' % indent)
        for i in range(self.rows):
            line = '%s    ' % indent
            for j in range(self.cols):
                line += '%d ' % self.interface.getValue(i,j)
            line += '\n'
            matrixSaveString.append(line)        
        matrixSaveString.append('%s  </coefficients>\n' % indent)
        if (self.mutes_interface != None):
            matrixSaveString.append('%s  <mutes>\n' % indent)
            for i in range(self.rows):
                line = '%s    ' % indent
                for j in range(self.cols):
                    line += '%d ' % self.mutes_interface.getValue(i,j)
                line += '\n'
                matrixSaveString.append(line)        
            matrixSaveString.append('%s  </mutes>\n' % indent)

        if (self.inverts_interface != None):
            matrixSaveString.append('%s  <inverts>\n' % indent)
            for i in range(self.rows):
                line = '%s    ' % indent
                for j in range(self.cols):
                    line += '%d ' % self.inverts_interface.getValue(i,j)
                line += '\n'
                matrixSaveString.append(line)        
            matrixSaveString.append('%s  </inverts>\n' % indent)

        return matrixSaveString

    def readSettings(self, readMatrixString, transpose_coeff):
        if readMatrixString[0].find("<row_number>") == -1:
            log.debug("Number of matrix rows must be specified")
            return False
        if readMatrixString[2].find("</row_number>") == -1:
            log.debug("Non-conformal xml file")
            return False
        n_rows = int(readMatrixString[1])

        if readMatrixString[3].find("<col_number>") == -1:
            log.debug("Number of matrix columns must be specified")
            return False
        if readMatrixString[5].find("</col_number>") == -1:
            log.debug("Non-conformal xml file")
            return False
        n_cols = int(readMatrixString[4])

        if transpose_coeff:
            if n_rows > self.cols:
                n_rows = self.cols
            if n_cols > self.rows:
                n_cols = self.rows
        else:
            if n_rows > self.rows:
                n_rows = self.rows
            if n_cols > self.cols:
                n_cols = self.cols
        log.debug("Setting %d rows and %d columns coefficients" % (n_rows, n_cols))

        try:
            idxb = readMatrixString.index('<coefficients>')
            idxe = readMatrixString.index('</coefficients>')
        except Exception:
            log.debug("No mixer matrix coefficients specified")
            idxb = -1
            idxe = -1
        if idxb >= 0:
            if idxe < idxb + n_rows + 1:
                log.debug("Incoherent number of rows in coefficients")
                return False
            i = 0
            for s in readMatrixString[idxb+1:idxb + n_rows + 1]:
                coeffs = s.split()
                if len(coeffs) < n_cols:
                    log.debug("Incoherent number of columns in coefficients")
                    return False
                j = 0
                for c in coeffs[0:n_cols]:
                    if transpose_coeff:
                        self.interface.setValue(j, i, int(c))
                    else:
                        self.interface.setValue(i, j, int(c))
                    j += 1
                i += 1
                del coeffs

        try:
            idxb = readMatrixString.index('<mutes>')
            idxe = readMatrixString.index('</mutes>')
        except Exception:
            log.debug("No mixer mute coefficients specified")
            idxb = -1
            idxe = -1
        if idxb >= 0:
            if idxe < idxb + n_rows + 1:
                log.debug("Incoherent number of rows in mute")
                return false
            i = 0
            for s in readMatrixString[idxb+1:idxb + n_rows + 1]:
                coeffs = s.split()
                if len(coeffs) < n_cols:
                    log.debug("Incoherent number of columns in mute")
                    return false
                j = 0
                for c in coeffs[0:n_cols]:
                    if transpose_coeff:
                        self.mutes_interface.setValue(j, i, int(c))
                    else:
                        self.mutes_interface.setValue(i, j, int(c))
                    j += 1
                i += 1
                del coeffs

        try:
            idxb = readMatrixString.index('<inverts>')
            idxe = readMatrixString.index('</inverts>')
        except Exception:
            log.debug("No mixer inverts coefficients specified")
            idxb = -1
            idxe = -1
        if idxb >= 0:
            if idxe < idxb + n_rows + 1:
                log.debug("Incoherent number of rows in inverts")
                return false
            i = 0
            for s in readMatrixString[idxb+1:idxb + n_rows + 1]:
                coeffs = s.split()
                if len(coeffs) < n_cols:
                    log.debug("Incoherent number of columns in inverts")
                    return false
                j = 0
                for c in coeffs[0:n_cols]:
                    if transpose_coeff:
                        self.inverts_interface.setValue(j, i, int(c))
                    else:
                        self.inverts_interface.setValue(i, j, int(c))
                    j += 1
                i += 1
                del coeffs

        self.refreshValues()
        return True

class VolumeSlider(QSlider):
    sliderChanged = pyqtSignal(tuple)
    def __init__(self, In, Out, value, parent):
        QSlider.__init__(self, QtCore.Qt.Vertical, parent)

        self.setTickPosition(QSlider.TicksBothSides)
        v_min = 10.0*toDBvalue(0)
        v_max = 10.0*toDBvalue(65536)
        self.setTickInterval(int((v_max-v_min)/10))
        self.setMinimum(int(v_min))
        self.setMaximum(int(v_max))
        self.setSingleStep(1)
        self.sliderSetValue(value)
        self.In = In
        self.Out = Out
        self.valueChanged.connect(self.sliderValueChanged)

    def sliderSetValue(self, value):
        #log.debug("Volume slider value changed( %i )" % value)
        v = 10.0*toDBvalue(value)
        #log.debug("Volume slider value changed(dB: %g )" % (0.1*v))
        self.setValue(int(v))

    def sliderReadValue(self, value):
        return fromDBvalue(0.1*value)

    # Restore absolute value from DB
    # Emit signal for further use, especially for matrix view
    def sliderValueChanged(self, value):
        value = fromDBvalue(0.1*value)
        self.sliderChanged.emit((self.In, self.Out, value))
        self.update()


class VolumeSliderValueInfo(QLineEdit):
    def __init__(self, In, Out, value, parent):
        QLineEdit.__init__(self, parent)

        self.setReadOnly(True)
        self.setAlignment(Qt.AlignCenter)
        self.setAutoFillBackground(True)
        self.setFrame(False)

        self.sliderSetMinimalDim()

        self.bgcolors = BckgrdColorForNumber()

        self.sliderSetValue(value)

    def sliderSetMinimalDim(self):
        fontmetrics = self.fontMetrics()
        self.setMinimumSize(fontmetrics.boundingRect("-00.0 dB").size()*1.1)
        
    def sliderSetValue(self, value):
        color = self.bgcolors.getColor(value)
        palette = self.palette()
        palette.setColor(QPalette.Active, QPalette.Base, color)
        palette.setColor(QPalette.Active, QPalette.Text, self.bgcolors.getFrgdColor(color))
        self.setPalette(palette)

        v = round(toDBvalue(value),1)
        if (v > -40):
            text = "%.1f dB" % v
        else:
            symb_inf = u"\u221E"
            text = "-" + symb_inf + " dB"

        self.setText(text)
        
class BalanceSlider(QSlider):
    sliderChanged = pyqtSignal(tuple)
    def __init__(self, In, Out, value, parent):
        QSlider.__init__(self, QtCore.Qt.Horizontal, parent)

        v_min = -50
        v_max = 50
        self.setTickPosition(QSlider.TicksBothSides)
        self.setTickInterval(int((v_max-v_min)/2))
        self.setMinimum(v_min)
        self.setMaximum(v_max)
        self.setSingleStep(1)
        self.In = In
        self.Out = Out
        self.sliderSetValue(value)
        self.valueChanged.connect(self.sliderValueChanged)

    def sliderSetValue(self, value):
        #log.debug("Balance fader value set( %d, %d, %f )" % (self.In, self.Out, value))
        v = int(round(50.0*value, 2))
        self.setValue(v)

    def sliderReadValue(self):
        return float(round(self.value()/50.0, 2))

    def sliderValueChanged(self, value):
        value = float(round(self.value()/50.0, 2))
        #log.debug("Balance fader value changed( %d, %d, %f )" % (self.In, self.Out, value))
        self.sliderChanged.emit((self.In, self.Out, value))

# Slider view widget
class SliderControlView(QWidget):
    valueChanged = pyqtSignal(tuple)
    def __init__(self, parent, servername, basepath, rule="Columns_are_inputs", shortname=False, shortinname="Ch", shortoutname="Ch", stereochannels = []):
        QWidget.__init__(self, parent)

        if not ffado.config.bypassdbus:
            self.bus = dbus.SessionBus()
            self.dev = self.bus.get_object(servername, basepath)
            self.interface = dbus.Interface(self.dev, dbus_interface="org.ffado.Control.Element.MatrixMixer")

        self.rule = rule
        self.shortname = shortname
        self.shortinname = shortinname
        self.shortoutname = shortoutname

        self.stereochannels = stereochannels

        self.out = []
        self.nbIn = self.getNbIn()
        self.nbOut = self.getNbOut()
        self.outmatrix = []

        k = 0
        for i in range(self.nbOut):
            widget = QWidget(parent)
            v_layout = QVBoxLayout(widget)
            v_layout.setAlignment(Qt.AlignCenter)
            widget.setLayout(v_layout)
            self.out.append(widget)

            self.out[i].is_stereo = False
            self.out[i].out_1 = k
            self.outmatrix.append(i)
            self.out[i].outname = "Out %d" % (k+1)
            if k in self.stereochannels:
                self.out[i].is_stereo = True
                self.out[i].outname += "+%d" % (k+2)
                self.outmatrix.append(i)

            self.out[i].lbl = []

            # Mixer/Out info label
            if (self.nbOut > 1):
                lbl = QLabel(widget)
                lbl.setText(self.getOutName(i, self.shortname))
                lbl.setAlignment(Qt.AlignCenter)
                v_layout.addWidget(lbl)
                self.out[i].lbl.append(lbl)

            h_layout_wid = QWidget(widget)
            h_layout = QHBoxLayout(h_layout_wid)
            h_layout.setAlignment(Qt.AlignCenter)
            h_layout_wid.setLayout(h_layout)
            v_layout.addWidget(h_layout_wid)
            self.out[i].volume = []
            self.out[i].svl = []
            self.out[i].balance = []

            for j in range(self.nbIn):
                h_v_layout_wid = QWidget(h_layout_wid)
                h_v_layout = QVBoxLayout(h_v_layout_wid)
                h_v_layout.setAlignment(Qt.AlignCenter)
                h_v_layout_wid.setLayout(h_v_layout)
                h_layout.addWidget(h_v_layout_wid)

                # Mixer/In info label
                if (self.nbIn > 1):
                    lbl = QLabel(h_v_layout_wid)
                    lbl.setText(self.getInName(j, self.shortname))
                    lbl.setAlignment(Qt.AlignCenter)
                    h_v_layout.addWidget(lbl)
                    self.out[i].lbl.append(lbl)

                h_v_h_layout_wid = QWidget(h_v_layout_wid)
                h_v_h_layout = QHBoxLayout(h_v_h_layout_wid)
                h_v_h_layout.setAlignment(Qt.AlignCenter)
                h_v_h_layout_wid.setLayout(h_v_h_layout)
                h_v_layout.addWidget(h_v_h_layout_wid)

                volume = VolumeSlider(j, i, self.getVolumeValue(j,i), h_v_h_layout_wid)
                h_v_h_layout.addWidget(volume)
                self.out[i].volume.append(volume)
                self.volumeConnect(volume)

                # Volume slider info
                svl = VolumeSliderValueInfo(j, i, self.getVolumeValue(j,i), h_v_layout_wid)
                h_v_layout.addWidget(svl)
                self.out[i].svl.append(svl)

                # Balance fader
                if self.out[i].is_stereo:
                    balance = BalanceSlider(j, i, self.getBalanceValue(j,i), h_v_layout_wid)
                    h_v_layout.addWidget(balance)
                    self.out[i].balance.append(balance)
                    self.balanceConnect(balance)
            k += 1
            if self.out[i].is_stereo:
                k += 1

    def volumeConnect(self, volume):
        volume.sliderChanged.connect(self.valueChangedVolume)

    def volumeDisconnect(self, volume):
        volume.sliderChanged.disconnect(self.valueChangedVolume)

    def balanceConnect(self, balance):
        balance.sliderChanged.connect(self.valueChangedBalance)

    def balanceDisconnect(self, balance):
        balance.sliderChanged.disconnect(self.valueChangedBalance)

    def getNbIn(self):
        if ffado.config.bypassdbus:
            return 2
        if (self.rule == "Columns_are_inputs"):
            return self.interface.getColCount()
        else:
            return self.interface.getRowCount()
        
    def getNbOut(self):
        if ffado.config.bypassdbus:
            return 2
        if (self.rule == "Columns_are_inputs"):
            nbout = self.interface.getRowCount()
        else:
            nbout = self.interface.getColCount()
        return nbout-len(self.stereochannels)
        
    def getVolumeValue(self, In, i):
        if ffado.config.bypassdbus:
            return 1
        Out = self.out[i].out_1
        if (self.rule == "Columns_are_inputs"):
            vl = self.interface.getValue(Out, In)           
        else:
            vl = self.interface.getValue(In, Out)
        if (self.out[i].is_stereo):
            if (self.rule == "Columns_are_inputs"):
                vr = self.interface.getValue(Out+1, In)           
            else:
                vr = self.interface.getValue(In, Out+1)
            return getStereoVolume(vl, vr)
        else:
            return vl;     

    def getBalanceValue(self, In, i):
        if ffado.config.bypassdbus:
            return 0.5
        Out = self.out[i].out_1
        if (self.rule == "Columns_are_inputs"):
            vl = self.interface.getValue(Out, In)           
            vr = self.interface.getValue(Out+1, In)           
        else:
            vl = self.interface.getValue(In, Out)
            vr = self.interface.getValue(In, Out+1)
        return getStereoBalance(vl, vr)

    def setValue(self, In, Out, val):
        if ffado.config.bypassdbus:
            return
        if (self.rule == "Columns_are_inputs"):
            return self.interface.setValue(Out, In, val)           
        else:
            return self.interface.setValue(In, Out, val)            

    def updateValues(self, n):
        nbitems = len(n) // 3
        for j in range(nbitems):
            n_0 = n[3*j]    
            n_1 = n[3*j+1]   
            n_2 = n[3*j+2] 
            i = self.outmatrix[n_1]
            if (self.out[i].is_stereo):
                v = self.getVolumeValue(n_0, i)
                self.volumeDisconnect(self.out[i].volume[n_0])
                self.balanceDisconnect(self.out[i].balance[n_0])
                self.out[i].volume[n_0].sliderSetValue(v)
                self.out[i].svl[n_0].sliderSetValue(v)
                b = self.getBalanceValue(n_0, i)       
                # log.debug("update Value (%d %d %d %f)" % (n_0, i, v, b))
                self.out[i].balance[n_0].sliderSetValue(b)
                self.volumeConnect(self.out[i].volume[n_0])
                self.balanceConnect(self.out[i].balance[n_0])
            else:
                v = n_2
                # log.debug("update Value (%d %d %d)" % (n_0, i, v))
                self.volumeDisconnect(self.out[i].volume[n_0])
                self.out[i].volume[n_0].sliderSetValue(v)
                self.out[i].svl[n_0].sliderSetValue(v)
                self.volumeConnect(self.out[i].volume[n_0])
        
    def valueChangedVolume(self, n):
        #log.debug("VolumeSlider.valueChanged( %s )" % str(n))
        v = n[2]
        n1 = self.out[n[1]].out_1
        if (self.out[n[1]].is_stereo):
            b = self.out[n[1]].balance[n[0]].value()/50.0
            vl = int(getVolumeLeft(v, b))
            self.setValue(n[0], n1, vl)
            n2 = n1+1
            vr = int(getVolumeRight(v, b))
            self.setValue(n[0], n2, vr)
            n_t = (n[0], n1, vl, n[0], n2, vr)
            self.valueChanged.emit(n_t)
        else:
            self.setValue(n[0], n1, v)
            n_t = (n[0], n1, v)
            self.valueChanged.emit(n_t)
        self.out[n[1]].svl[n[0]].sliderSetValue(v)

    def valueChangedBalance(self, n):
        #log.debug("BalanceSlider.valueChanged( %s )" % str(n))
        n1 = self.out[n[1]].out_1
        v = fromDBvalue(0.1*self.out[n[1]].volume[n[0]].value())
        b = n[2]
        vl = int(getVolumeLeft(v, b))
        self.setValue(n[0], n1, vl)
        n2 = n1+1
        vr = int(getVolumeRight(v, b))
        self.setValue(n[0], n2, vr)
        n_t = (n[0], n1, vl, n[0], n2, vr)
        self.valueChanged.emit(n_t)

    def getOutName(self, i, shortname):
        self.shortname = shortname
        k = self.out[i].out_1
        if (shortname):
            if (self.out[i].is_stereo):
                number = " %d+%d" % (k+1, k+2)
            else:
                number = " %d" % (k+1)
            name = self.shortoutname + number
            return name
        else:
            if ffado.config.bypassdbus:
                return 'OutName ' + str(i)
            if (self.rule == "Columns_are_inputs"):                
                if (self.out[i].is_stereo):
                    name = self.interface.getRowName(k).replace('\n','')+" + "+self.interface.getRowName(k+1).replace('\n','')
                else:
                    name = self.interface.getRowName(k).replace('\n','')
                return name
            else:
                if (self.out[i].is_stereo):
                    name = self.interface.getColName(k).replace('\n','')+" + "+self.interface.getColName(k+1).replace('\n','')
                else:
                    name = self.interface.getColName(k).replace('\n','')
                return name

    def getInName(self, j, shortname):
        self.shortname = shortname
        if (shortname):
            number = " %d" % (j+1)
            name = self.shortinname + number
            return name
        else:
            if ffado.config.bypassdbus:
                return 'InName ' + str(j)
            if (self.rule == "Columns_are_inputs"):
                return self.interface.getColName(j)            
            else:
                return self.interface.getRowName(j)            

    # Update when routing is modified
    def updateRouting(self):
        for i in range(self.nbOut):
            if (self.nbOut > 1):
                last_name = self.out[i].lbl[0].text()
                out_name = self.getOutName(i, self.shortname)
                if last_name != out_name:
                    #log.debug("SliderControlView.updateRouting( %s )" % str(out_name))
                    self.out[i].lbl[0].setText(out_name)
            if (self.nbIn > 1):
                for j in range(self.nbIn):
                    last_name = self.out[i].lbl[j+1].text()
                    in_name = self.getInName(j, self.shortname)
                    if last_name != in_name:
                        #log.debug("SliderControlView.updateRouting( %s )" % str(in_name))
                        self.out[i].lbl[j+1].setText(in_name)

    def refreshValues(self):
        for n_out in range(self.nbOut):
            for n_in in range(self.nbIn):
                i = self.outmatrix[n_out]
                v = self.getVolumeValue(n_in, i)
                if (self.out[i].is_stereo):
                    self.volumeDisconnect(self.out[i].volume[n_in])
                    self.out[i].volume[n_in].sliderSetValue(v)
                    self.out[i].svl[n_in].sliderSetValue(v)
                    b = self.getBalanceValue(n_in, i)       
                    # log.debug("update Value (%d %d %d %f)" % (n_0, i, v, b))
                    self.out[i].balance[n_in].sliderSetValue(b)
                    self.volumeConnect(self.out[i].volume[n_in])
                else:
                    # log.debug("update Value (%d %d %d)" % (n_0, i, v))
                    self.out[i].volume[n_in].sliderSetValue(v)
                    self.out[i].svl[n_in].sliderSetValue(v)

    def saveSettings(self, indent):
        if ffado.config.bypassdbus:
            rows = 2
            cols = 2
        else:
            rows = self.interface.getRowCount()
            cols = self.interface.getColCount()
        matrixSaveString = []
        matrixSaveString.append('%s  <row_number>\n' % indent)
        matrixSaveString.append('%s    %d\n' % (indent, rows))        
        matrixSaveString.append('%s  </row_number>\n' % indent)
        matrixSaveString.append('%s  <col_number>\n' % indent)
        matrixSaveString.append('%s    %d\n' % (indent, cols))        
        matrixSaveString.append('%s  </col_number>\n' % indent)
        matrixSaveString.append('%s  <coefficients>\n' % indent)
        for i in range(rows):
            line = '%s    ' % indent
            for j in range(cols):
                line += '%d ' % self.interface.getValue(i,j)
            line += '\n'
            matrixSaveString.append(line)        
        matrixSaveString.append('%s  </coefficients>\n' % indent)

        return matrixSaveString

    def readSettings(self, readMatrixString, transpose_coeff):
        if ffado.config.bypassdbus:
            rows = 2
            cols = 2
        else:
            rows = self.interface.getRowCount()
            cols = self.interface.getColCount()
        if readMatrixString[0].find("<row_number>") == -1:
            log.debug("Number of matrix rows must be specified")
            return False
        if readMatrixString[2].find("</row_number>") == -1:
            log.debug("Non-conformal xml file")
            return False
        n_rows = int(readMatrixString[1])

        if readMatrixString[3].find("<col_number>") == -1:
            log.debug("Number of matrix columns must be specified")
            return False
        if readMatrixString[5].find("</col_number>") == -1:
            log.debug("Non-conformal xml file")
            return False
        n_cols = int(readMatrixString[4])

        if transpose_coeff:
            if n_rows > cols:
                n_rows = cols
            if n_cols > rows:
                n_cols = rows
        else:
            if n_rows > rows:
                n_rows = rows
            if n_cols > cols:
                n_cols = cols
        log.debug("Setting %d rows and %d columns coefficients" % (n_rows, n_cols))

        try:
            idxb = readMatrixString.index('<coefficients>')
            idxe = readMatrixString.index('</coefficients>')
        except Exception:
            log.debug("No mixer matrix coefficients specified")
            idxb = -1
            idxe = -1
        if idxb >= 0:
            if idxe < idxb + n_rows + 1:
                log.debug("Incoherent number of rows in coefficients")
                return False
            i = 0
            for s in readMatrixString[idxb+1:idxb + n_rows + 1]:
                coeffs = s.split()
                if len(coeffs) < n_cols:
                    log.debug("Incoherent number of columns in coefficients")
                    return False
                j = 0
                for c in coeffs[0:n_cols]:
                    if transpose_coeff:
                        self.interface.setValue(j, i, int(c))
                    else:
                        self.interface.setValue(i, j, int(c))
                    j += 1
                i += 1
                del coeffs

        self.refreshValues()
        return True

from functools import partial

class MatrixMixer(QWidget):
    def __init__(self, servername, basepath, parent=None, rule="Columns_are_inputs", sliderMaxValue=-1, mutespath=None, invertspath=None, smallFont=False, taborientation=QTabWidget.West, tabshape=QTabWidget.Triangular):
        QWidget.__init__(self, parent)
        self.servername = servername
        self.basepath = basepath
        self.sliderMaxValue = sliderMaxValue
        self.mutespath = mutespath
        self.invertspath = invertspath
        self.smallFont = smallFont

        self.layout = QVBoxLayout(self)
        self.setLayout(self.layout)

        # Mixer view Tool bar
        mxv_set = QToolBar("View settings", self)

        # Here is a hack; the first action button appears to behaves strangely,
        # possibly a PyQt bug (or an unsufficient fair implementation of it)
        # Feel free to remove the next three lines at a time in the future
        hack = QAction(" ", mxv_set)
        hack.setDisabled(True)
        mxv_set.addAction(hack)

        transpose_matrix = QAction("Transpose", mxv_set)
        self.transpose = False
        transpose_matrix.setShortcut('Ctrl+T')
        transpose_matrix.setToolTip("Invert rows and columns in Matrix view")
        mxv_set.addAction(transpose_matrix)
        transpose_matrix.triggered.connect(self.transposeMatrixView)
        mxv_set.addSeparator()

        self.hide_matrix = QAction("Hide matrix", mxv_set)
        self.hide_matrix_bool = False
        mxv_set.addAction(self.hide_matrix)
        self.hide_matrix.triggered.connect(self.hideMatrixView)
        mxv_set.addSeparator()

        self.hide_per_output = QAction("Hide per Output", mxv_set)
        self.hide_per_output_bool = False
        mxv_set.addAction(self.hide_per_output)
        self.hide_per_output.triggered.connect(self.hidePerOutputView)
        mxv_set.addSeparator()

        self.use_short_names = QAction("Short names", mxv_set)
        self.short_names_bool = False
        mxv_set.addAction(self.use_short_names)
        self.use_short_names.setToolTip("Use short or full names for input and output channels")
        self.use_short_names.triggered.connect(self.shortChannelNames)
        mxv_set.addSeparator()

        font_switch_lbl = QLabel(mxv_set)
        font_switch_lbl.setText("Font size ")
        mxv_set.addWidget(font_switch_lbl)
        font_switch = QComboBox(mxv_set)
        font_switch.setToolTip("Labels font size")
        font = font_switch.font()
        for i in range(10):
            font_switch.addItem(" %d " % (font.pointSize()+4-i))
        font_switch.setCurrentIndex(font_switch.findText(" %d " % font.pointSize()))
        mxv_set.addWidget(font_switch)
        mxv_set.addSeparator()
        font_switch.activated.connect(self.changeFontSize)

        self.layout.addWidget(mxv_set)
        self.mxv_set = mxv_set

        # First tab is for matrix view
        # Next are for "per Out" view
        self.tabs = QTabWidget(self)
        self.tabs.setTabPosition(taborientation)
        self.tabs.setTabShape(tabshape)
        self.layout.addWidget(self.tabs)

        # Inputs/Outputs versus rows/columns rule
        self.rule = rule

        # Matrix view tab
        if (rule == "Columns_are_inputs"):
            self.matrix = MatrixControlView(servername, basepath, self, sliderMaxValue, mutespath, invertspath, smallFont, self.short_names_bool, "In", "Out", self.transpose)
        else:
            self.matrix = MatrixControlView(servername, basepath, self, sliderMaxValue, mutespath, invertspath, smallFont, self.short_names_bool, "Out", "In", self.transpose)
        self.matrix.valueChanged.connect(self.matrixControlChanged)

        self.scrollarea_matrix = QScrollArea(self.tabs)
        self.scrollarea_matrix.setWidgetResizable(True)
        self.scrollarea_matrix.setWidget(self.matrix)
        self.tabs.addTab(self.scrollarea_matrix, " Matrix ")

        # Add stereo/mono output choice in tool bar
        if (rule == "Columns_are_inputs"):
            if (self.transpose):
                nb_out_mono = self.matrix.cols
            else:
                nb_out_mono = self.matrix.rows
        else:
            if (self.transpose):
                nb_out_mono = self.matrix.rows
            else:
                nb_out_mono = self.matrix.cols

        stereo_switch_lbl = QLabel(mxv_set)
        stereo_switch_lbl.setText("Stereo: ")
        mxv_set.addWidget(stereo_switch_lbl)

        self.stereo_channels = []

        self.stereo_switch = []
        for i in range(int(nb_out_mono/2)):
            stereo_switch = QPushButton("%d+%d" % (2*i+1, 2*i+2), mxv_set)
            stereo_switch.setToolTip("Set these output channels as stereo")
            stereo_switch.setCheckable(True)
            stereo_switch.clicked.connect(partial(self.switchStereoChannel, i))
            stereo_switch.setMinimumSize(stereo_switch_lbl.fontMetrics().boundingRect("%d+%d" % (nb_out_mono, nb_out_mono)).size()*1.05)
            stereo_switch.setSizePolicy(QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum))
            stereo_switch.is_stereo = False
            mxv_set.addWidget(stereo_switch)
            self.stereo_switch.append(stereo_switch)
        mxv_set.addSeparator()

        # Per out view tabs
        self.perOut = SliderControlView(self, servername, basepath, rule, self.short_names_bool, "In", "Out", self.stereo_channels)
        self.perOut.valueChanged.connect(self.sliderControlChanged)
        for i in range(self.perOut.nbOut):
            self.perOut.out[i].scrollarea = QScrollArea(self.tabs)
            self.perOut.out[i].scrollarea.setWidgetResizable(True)
            self.perOut.out[i].scrollarea.setWidget(self.perOut.out[i])
            self.tabs.addTab(self.perOut.out[i].scrollarea, " %s " % self.perOut.out[i].outname)

    def transposeMatrixView(self):
        self.transpose = not(self.transpose)
        self.tabs.removeTab(0)
        self.scrollarea_matrix.destroy()
        if (self.rule == "Columns_are_inputs"):
            self.matrix = MatrixControlView(self.servername, self.basepath, self, self.sliderMaxValue, self.mutespath, self.invertspath, self.smallFont, self.short_names_bool, "In", "Out", self.transpose)
        else:
            self.matrix = MatrixControlView(self.servername, self.basepath, self, self.sliderMaxValue, self.mutespath, self.invertspath, self.smallFont, self.short_names_bool, "Out", "In", self.transpose)
        self.matrix.valueChanged.connect(self.matrixControlChanged)

        self.scrollarea_matrix = QScrollArea(self.tabs)
        self.scrollarea_matrix.setWidgetResizable(True)
        self.scrollarea_matrix.setWidget(self.matrix)
        self.tabs.insertTab(0, self.scrollarea_matrix, "Matrix")
        self.tabs.setCurrentIndex(0)
        
    def hideMatrixView(self):
        self.hide_matrix_bool = not(self.hide_matrix_bool)
        if (self.hide_matrix_bool):
            self.tabs.removeTab(0)
            self.hide_matrix.setText("Show Matrix")
        else:
            self.tabs.insertTab(0, self.scrollarea_matrix, "Matrix")
            self.tabs.setCurrentIndex(0)
            self.hide_matrix.setText("Hide Matrix")
            
    def hidePerOutputView(self):
        self.hide_per_output_bool = not(self.hide_per_output_bool)
        if (self.hide_per_output_bool):
            index_0 = 1
            if (self.hide_matrix_bool):
                index_0 = 0
            for i in range(self.perOut.nbOut):
                self.tabs.removeTab(index_0)
            self.hide_per_output.setText("Show per Output")
        else:
            for i in range(self.perOut.nbOut):
                self.tabs.insertTab(i+1, self.perOut.out[i].scrollarea, " %s " % self.perOut.out[i].outname)
            self.hide_per_output.setText("Hide per Output")

    # Font size for channel names
    def changeFontSize(self, size):
        font = self.mxv_set.font()
        font.setPointSize(int(size))
        self.mxv_set.setFont(font)

        font = self.tabs.font()
        font.setPointSize(int(size))
        self.tabs.setFont(font)

        font = self.matrix.font()
        font.setPointSize(int(size))
        self.matrix.setFont(font)

        font = self.perOut.font()
        font.setPointSize(int(size))
        self.perOut.setFont(font)

        for i in range(self.perOut.nbOut):
            for j in range(self.perOut.nbIn):
                self.perOut.out[i].svl[j].sliderSetMinimalDim()

    # Allows long name for Mixer/Out and /In to be hidden 
    def shortChannelNames(self):
        checked = not(self.short_names_bool)
        if (self.matrix.cols > 1):
            for i in range(self.matrix.cols):
                self.matrix.columnHeaders[i].name = self.matrix.getColName(i, checked)
                self.matrix.columnHeaders[i].lbl.setText(self.matrix.columnHeaders[i].name)

        if (self.matrix.rows > 1):
            for j in range(self.matrix.rows):
                self.matrix.rowHeaders[j].name = self.matrix.getRowName(j, checked)
                self.matrix.rowHeaders[j].lbl.setText(self.matrix.rowHeaders[j].name)       

        for i in range(self.perOut.nbOut):
            if (self.perOut.nbOut > 1):
                self.perOut.out[i].lbl[0].setText(self.perOut.getOutName(i, checked))
            if (self.perOut.nbIn > 1):
                for j in range(self.perOut.nbIn):
                    self.perOut.out[i].lbl[j+1].setText(self.perOut.getInName(j, checked))

        # Care for hidden columns
        if (self.matrix.cols > 1):
            for i in self.matrix.hiddenCols:
                self.matrix.columnHeaders[i].lbl.setText("%d" % (i+1))
        # Care for hidden rows
        if (self.matrix.rows > 1):
            for j in self.matrix.hiddenRows:
                self.matrix.rowHeaders[j].lbl.setText("%d" % (j+1))

        self.short_names_bool = checked
        if (self.short_names_bool):
            self.use_short_names.setText("Long names")
        else:
            self.use_short_names.setText("Short names")

    # Sliders value change
    #   Care that some recursive process is involved and only stop when exactly same values are involved
    # Matrix view
    def matrixControlChanged(self, n):
        # Update value needed for "per Out" view
        #log.debug("Update per Output( %s )" % str(n))
        nbitems = len(n) // 3
        if (self.rule == "Columns_are_inputs"):
           n_t = n
        else:
            n_t = ()
            for i in range(nbitems):
                n_t += (n[3*i+1], n[3*i], n[3*i+2])

        self.perOut.updateValues(n_t)

    # "per Out" view
    def sliderControlChanged(self, n):
        # Update value needed for matrix view
        #log.debug("Update Matrix( %s )" % str(n))
        nbitems = len(n) // 3
        if (((self.rule == "Columns_are_inputs") and not self.transpose) or ((self.rule != "Columns_are_inputs") and self.transpose)):
            n_t = ()
            for i in range(nbitems):
                n_t += (n[3*i+1], n[3*i], n[3*i+2])
        else:
            n_t = n

        self.matrix.updateValues(n_t)

    def refreshValues(self):
        # Refreshing matrix coefficient should be sufficient,
        #  propagating the changes to perOut view
        self.matrix.refreshValues()

    def switchStereoChannel(self, channel, is_stereo):
        #log.debug(" switching channels %d+%d to stereo/mono" % (2*channel, 2*channel+1))
        self.stereo_switch[channel].is_stereo = self.stereo_switch[channel].isChecked();
        if (self.stereo_switch[channel].is_stereo):
            self.stereo_channels.append(2*channel)
        else:
            self.stereo_channels.remove(2*channel)

        # tab 0 is for matrix except if it is hidden
        index_0 = 1
        if (self.hide_matrix_bool):
            index_0 = 0
        for i in range(self.perOut.nbOut):
            self.tabs.removeTab(index_0)
        self.perOut.destroy()
        self.perOut = SliderControlView(self, self.servername, self.basepath, self.rule, self.short_names_bool, "In", "Out", self.stereo_channels)
        self.perOut.valueChanged.connect(self.sliderControlChanged)
        current = 0
        for i in range(self.perOut.nbOut):
            self.perOut.out[i].scrollarea = QScrollArea(self.tabs)
            self.perOut.out[i].scrollarea.setWidgetResizable(True)
            self.perOut.out[i].scrollarea.setWidget(self.perOut.out[i])
            self.tabs.addTab(self.perOut.out[i].scrollarea, " %s " % self.perOut.out[i].outname)
            if self.perOut.out[i].out_1 == 2*channel:
                current = i

        self.tabs.setCurrentWidget(self.perOut.out[current].scrollarea)

    # Update when routing is modified
    def updateRouting(self):
        self.matrix.updateRouting()
        self.perOut.updateRouting()
        
    def saveSettings(self, indent):
        mixerString = []
        mixerString.append("%s<matrices>\n" % indent)
        mixerString.extend(self.matrix.saveSettings(indent))
        mixerString.append("%s</matrices>\n" % indent)
        mixerString.append("%s<stereo_outputs>\n" % indent)
        mixerString.append("%s  <number>\n" % indent)
        n = len(self.stereo_channels)
        mixerString.append("%s    %d\n" % (indent, n))
        mixerString.append("%s  </number>\n" % indent)
        if n > 0:
            mixerString.append("%s  <channels>\n" % indent)
            for i in self.stereo_channels:
                mixerString.append("%s    %d %d\n" % (indent, i+1, i+2))
            mixerString.append("%s  </channels>\n" % indent)
        mixerString.append("%s</stereo_outputs>\n" % indent)
        return mixerString

    def readSettings(self, readMixerString, transpose_coeff):
        try:
            idxb = readMixerString.index('<matrices>')
            idxe = readMixerString.index('</matrices>')
        except Exception:
            log.debug("No matrices found")
            idxb = -1
            idxe = -1
        if idxb >= 0:
            if idxe > idxb+1:
                readString = []
                for s in readMixerString[idxb+1:idxe]:
                    readString.append(s)
                if self.matrix.readSettings(readString, transpose_coeff):
                    log.debug("Mixer matrices settings modified")
                del readString
        try:
            idx = readMixerString.index('<stereo_outputs>')
        except Exception:
            log.debug("No stereo outputs channels information found")
            idx = -1
        if idx >= 0:
            if readMixerString[idx+1].find('<number>') == -1:
                log.debug("Number of stereo output channels must be specified")
                return False
            n = int(readMixerString[idx+2])
            if n > self.perOut.nbOut // 2:
                log.debug("Incoherent number of stereo channels")
                return False
            if n > 0:
                if readMixerString[idx+3].find('</number>') == -1:
                    log.debug("No </number> tag found")
                    return False
                if readMixerString[idx+4].find('<channels>') == -1:
                    log.debug("No <channels> tag found")
                    return False
                for s in readMixerString[idx+5:idx+5+n]:
                    i = (int(s.split()[0]) - 1)/2
                    self.stereo_switch[i].setChecked(True);
                    self.switchStereoChannel(i, True)
        return True
                    
#
# vim: et ts=4 sw=4 fileencoding=utf8



#
# Copyright (C) 2005-2008 by Daniel Wagner
#
# This file is part of FFADO
# FFADO = Free FireWire (pro-)audio drivers for Linux
#
# FFADO is based upon FreeBoB.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# from PyQt4.QtGui import QWidget
from ffado.import_pyqt import *

from ffado.config import *

import logging
log = logging.getLogger('edirolfa66')

class EdirolFa66Control(QWidget):
    def __init__(self, parent = None):
        QWidget.__init__(self, parent)
        uicLoad("ffado/mixer/edirolfa66", self)

        self.VolumeControls = {
            #          feature name, channel, qt slider
            'vol1'  :   ['/Mixer/Feature_Volume_1', 1, self.sldInput1],
            'vol2'  :   ['/Mixer/Feature_Volume_1', 2, self.sldInput2],
            'vol3'  :   ['/Mixer/Feature_Volume_2', 1, self.sldInput3],
            'vol4'  :   ['/Mixer/Feature_Volume_2', 2, self.sldInput4],
            'vol5'  :   ['/Mixer/Feature_Volume_3', 1, self.sldInput5],
            'vol6'  :   ['/Mixer/Feature_Volume_3', 2, self.sldInput6],

            'bal1'  :   ['/Mixer/Feature_LRBalance_1', 1, self.sldBal1],
            'bal2'  :   ['/Mixer/Feature_LRBalance_1', 2, self.sldBal2],
            'bal3'  :   ['/Mixer/Feature_LRBalance_2', 1, self.sldBal3],
            'bal4'  :   ['/Mixer/Feature_LRBalance_2', 2, self.sldBal4],
            'bal5'  :   ['/Mixer/Feature_LRBalance_3', 1, self.sldBal5],
            'bal6'  :   ['/Mixer/Feature_LRBalance_3', 2, self.sldBal6],
            }

    def setVolumeIn1(self, vol):
        self.setValue('vol1', vol)

    def setVolumeIn2(self, vol):
        self.setValue('vol2', vol)

    def setVolumeIn3(self, vol):
        self.setValue('vol3', vol)

    def setVolumeIn4(self, vol):
        self.setValue('vol4', vol)

    def setVolumeIn5(self, vol):
        self.setValue('vol5', vol)

    def setVolumeIn6(self, vol):
        self.setValue('vol6', vol)

    def setBalanceIn1(self, bal):
        self.setValue('bal1', bal)

    def setBalanceIn2(self, bal):
        self.setValue('bal2', bal)

    def setBalanceIn3(self, bal):
        self.setValue('bal3', bal)

    def setBalanceIn4(self, bal):
        self.setValue('bal4', bal)

    def setBalanceIn5(self, bal):
        self.setValue('bal5', bal)

    def setBalanceIn6(self, bal):
        self.setValue('bal6', bal)

    def setValue(self, name, val):
        ctrl = self.VolumeControls[name]
        log.debug("setting %s to %d" % (name, val))
        self.hw.setContignuous(ctrl[0], val, idx = ctrl[1])

    def initValues(self):
        for name, ctrl in self.VolumeControls.items():
            val = self.hw.getContignuous(ctrl[0], idx = ctrl[1])
            log.debug("%s value is %d" % (name , val))

            # Workaround: The current value is not properly initialized
            # on the device and returns after bootup always 0.
            # Though we happen to know what the correct value should
            # be therefore we overwrite the 0 
            if name[0:3] == 'bal' and val == 0:
                if ctrl[1] == 1:
                    val = 32512
                else:
                    val = -32768

            ctrl[2].setValue(val)

# vim: et



# -*- coding: utf-8 -*-
REGISTER_URL = 'http://ffado.org/deviceregistration/register.php?action=register'
INI_FILE_PATH = "~/.ffado/registration.ini"

FFADO_CONFIG_DIR = "~/.ffado"

FFADO_VERSION="2.4.9-"

FFADO_DBUS_SERVER = 'org.ffado.Control'
FFADO_DBUS_BASEPATH = '/org/ffado/Control'

POLL_SLEEP_TIME_MSEC = 100 # 100ms

PYTHONDIR = "/usr/lib/python3.13/site-packages"
SHAREDIR = "/usr/share/libffado"

USER_CONFIG_FILE = "~/.ffado/configuration"
SYSTEM_CONFIG_FILE = "/usr/share/libffado/configuration"

DEBUG = False

# If set true it will open all mixers
bypassdbus = False

UIDIR = PYTHONDIR
import os.path
if os.path.exists('support/mixer-qt4/ffado/mixer/globalmixer.ui'):
    UIDIR = "support/mixer-qt4"
if os.path.exists('ffado/mixer/globalmixer.ui'):
    UIDIR = "."

import os.path

# from PyQt4 import uic
from ffado.import_pyqt import *

def uicLoad(file, object):
    if not file.endswith(".ui"):
        file += ".ui"
    uic.loadUi(os.path.join(UIDIR,file), object)

# vim: et



#
# Copyright (C) 2008 by Arnold Krille
#               2013 by Philippe Carriere
#
# This file is part of FFADO
# FFADO = Free FireWire (pro-)audio drivers for Linux
#
# FFADO is based upon FreeBoB.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# from PyQt4.QtCore import QObject, pyqtSlot
# from PyQt4.QtGui import QWidget, QMessageBox
from ffado.import_pyqt import *

from ffado.config import *

import logging
log = logging.getLogger('global')

class GlobalMixer(QWidget):
    def __init__(self, parent, name=None):
        QWidget.__init__(self, parent)
        uicLoad("ffado/mixer/globalmixer", self)
        self.setName(name)

    def setName(self,name):
        if name is not None:
            self.lblName.setText(name)
            self.lblName.show()
        else:
            self.lblName.hide()

    @pyqtSlot(int)
    def on_clocksource_activated( self, clock ):
        #log.debug("updateClockSource( " + str(clock) + " )")
        if self.clockselect.canChangeValue():
            self.clockselect.select( clock )
        else:
            msg = QMessageBox()
            msg.question( msg, "Error", \
                "<qt>Clock source change not permitted. Is streaming active?</qt>", \
                QMessageBox.Ok )
            self.clocksource.setEnabled(False)
            return

        selected = self.clockselect.selected()
        if selected != clock:
            clockname = self.clockselect.getEnumLabel( clock )
            msg = QMessageBox()
            msg.question( msg, "Failed to select clock source", \
                "<qt>Could not select %s as clock source.</qt>" % clockname, \
                QMessageBox.Ok )
            self.clocksource.setCurrentIndex( selected )

    @pyqtSlot(int)
    def on_samplerate_activated( self, sr ):
        log.debug("on_samplerate_activated( " + str(sr) + " )")
        # If there's no clock, don't bother trying to set the sample rate
        if (self.no_clock == 1):
            return
        if self.samplerateselect.canChangeValue():
            self.samplerateselect.select( sr )
            if 'onSamplerateChange' in dir(self):
                self.onSamplerateChange()
                log.debug("Mixer configuration updated")
        else:
            msg = QMessageBox()
            msg.question( msg, "Error", \
                "<qt>Sample rate change not permitted. Is streaming active?</qt>", \
                QMessageBox.Ok )
            self.samplerate.setEnabled(False)
            return

        selected = self.samplerateselect.selected()
        if selected != sr:
            srname = self.samplerateselect.getEnumLabel( sr )
            msg = QMessageBox()
            msg.question( msg, "Failed to select sample rate", \
                "<qt>Could not select %s as samplerate.</qt>" % srname, \
                QMessageBox.Ok )
            self.samplerate.setCurrentIndex( selected )

    @pyqtSlot()
    def on_txtNickname_returnPressed( self ):
        if self.nickname.canChangeValue():
            asciiData = self.txtNickname.text().toAscii()
            self.nickname.setText( asciiData.data() )
        else:
            self.txtNickname.setText( self.nickname.text() )

    def refreshSampleRates( self ):
        no_clock_status = 0
        n_rates = self.samplerateselect.count()
        if (n_rates<1 or self.samplerateselect.getEnumLabel(0)=="0"):
            no_clock_status = 1;

        # Except for changes to the "no clock" status (where the number of
        # "frequencies" can remain at 1), the following test won't account
        # for cases where the frequency list changes but the total number of
        # frequencies remains the same.  If a device comes along for which
        # this is a problem, an alternative approach will be needed.
        if (no_clock_status!=self.no_clock or n_rates!=self.num_rates):
            self.no_clock = 0;
            self.num_rates = n_rates
            self.samplerate.clear()
            for i in range( self.num_rates ):
                label = self.samplerateselect.getEnumLabel( i )
                if (label == "0"):
                    label = "No clock found";
                    self.no_clock = 1;
                self.samplerate.insertItem( self.num_rates, label )
            if (self.no_clock != 1):
                self.samplerate.setCurrentIndex( self.samplerateselect.selected() )
            else:
                self.samplerate.setCurrentIndex(0);

    def initValues( self ):
        #print( "GlobalMixer::initValues()" )
        nb_clocks = self.clockselect.count()
        for i in range( nb_clocks ):
            self.clocksource.insertItem( nb_clocks, self.clockselect.getEnumLabel( i ) )
        self.clocksource.setCurrentIndex( self.clockselect.selected() )

        self.no_clock = 0;
        self.num_rates = -1;
        self.refreshSampleRates();

        self.txtNickname.setText( self.nickname.text() )

        self.samplerate.setEnabled(self.samplerateselect.canChangeValue())
        self.clocksource.setEnabled(self.clockselect.canChangeValue())
        if self.nickname.canChangeValue():
            self.txtNickname.setEnabled(True)
        else:
            self.txtNickname.setEnabled(False)

        self.streaming_status = self.streamingstatus.selected()

    def polledUpdate(self):
        self.samplerate.setEnabled(self.samplerateselect.canChangeValue())
        self.clocksource.setEnabled(self.clockselect.canChangeValue())
        self.txtNickname.setEnabled(self.nickname.canChangeValue())
        ss = self.streamingstatus.selected()
        ss_txt = self.streamingstatus.getEnumLabel(ss)
        if ss_txt == 'Idle':
            self.chkStreamIn.setChecked(False)
            self.chkStreamOut.setChecked(False)
        elif ss_txt == 'Sending':
            self.chkStreamIn.setChecked(False)
            self.chkStreamOut.setChecked(True)
        elif ss_txt == 'Receiving':
            self.chkStreamIn.setChecked(True)
            self.chkStreamOut.setChecked(False)
        elif ss_txt == 'Both':
            self.chkStreamIn.setChecked(True)
            self.chkStreamOut.setChecked(True)

        if (ss!=self.streaming_status and ss_txt!='Idle'):
            sr = self.samplerate.currentIndex()
            self.samplerate.setCurrentIndex( self.samplerateselect.selected() )
            # Check (and update) if device configuration needs to be updated
            if ( self.samplerateselect.devConfigChanged(sr) ):
                log.debug("Samplerate modified by external client")
                if 'onSamplerateChange' in dir(self):
                    self.onSamplerateChange()
                    log.debug("Mixer configuration updated ")
                else:
                    log.debug("Mixer configuration need to be updated but no means is provided")
        self.streaming_status = ss

        # Allow for devices whose sample rates can change dynamically (for
        # example, in response to changes in external clock frequency)
        self.refreshSampleRates();

    def saveSettings(self, indent):
        saveString = []
        saveString.append('%s<nickname>\n' % indent)
        saveString.append('%s  ' % indent + str(self.txtNickname.text()) + '\n')
        saveString.append('%s</nickname>\n' % indent)
        saveString.append('%s<clock>\n' % indent)
        saveString.append('%s  ' % indent + str(self.clockselect.getEnumLabel(self.clockselect.selected())) + '\n')
        saveString.append('%s</clock>\n' % indent)
        saveString.append('%s<samplerate>\n' % indent)
        saveString.append('%s  ' % indent + str(self.samplerateselect.getEnumLabel(self.samplerateselect.selected())) + ' \n')
        saveString.append('%s</samplerate>\n' % indent)
        return saveString

    def readSettings(self, readString):
        # Nickname
        try:
            idx = readString.index('<nickname>')
        except Exception:
            print( "No nickname found" )
            idx = -1
        if idx >= 0:
            nickname = readString[idx+1]
            if self.nickname.canChangeValue():
                self.txtNickname.setText(nickname)
                self.on_txtNickname_returnPressed()
            log.debug("Nickname changed for %s" % nickname)
        # Clock
        try:
            idx = readString.index('<clock>')
        except Exception:
            print( "No clock found" )
            idx = -1
        if idx >= 0:
            clock = readString[idx+1]
            nb_clocks = self.clockselect.count()
            clockLabel = []
            for i in range( nb_clocks ):
                clockLabel.append(self.clockselect.getEnumLabel(i))
            try:
                idxclock = clockLabel.index(clock)
            except Exception:
                print( "No %s clock found" % clock )
                idxclock = -1
            if idxclock >= 0:
                self.on_clocksource_activated(idxclock)           
                self.clocksource.setCurrentIndex(self.clockselect.selected())
                log.debug("Clock set to index %d (%s)" % (idxclock, clock))
            del clockLabel
        # Samplerate
        try:
            idx = readString.index('<samplerate>')
        except Exception:
            print( "Samplerate not found" )
            idx = -1
        if idx >= 0:
            samplerate = readString[idx+1]
            nb_srate = self.samplerateselect.count()
            srateLabel = []
            for i in range( nb_srate ):
              srateLabel.append(self.samplerateselect.getEnumLabel(i))
            try:
                idxsrate = srateLabel.index(samplerate)
            except Exception:
                print( "No %s samplerate found" % samplerate )
                idxsrate = -1
            if idxsrate >= 0:
                self.on_samplerate_activated(idxsrate)    
                self.samplerate.setCurrentIndex(self.samplerateselect.selected())
                log.debug("Samplerate set to index %d (%s)" % (idxsrate, samplerate))
         
# vim: et



#
# presonus_firebox.py - Qt4/FFADO application for Presonus FIREBOX
# Copyright (C) 2013 Takashi Sakamoto <o-takashi@sakamocchi.jp>
#
# This file is part of FFADO
# FFADO = Free FireWire (pro-)audio drivers for Linux
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# from PyQt4 import QtGui, QtCore
# from PyQt4.QtCore import Qt
# from PyQt4.QtGui import QGridLayout
# from PyQt4.QtGui import QWidget, QLabel, QSizePolicy, QToolButton, QSlider, QDial
from ffado.import_pyqt import *

from math import log10
from ffado.config import *

class Presonus_FP10(QWidget):
    outputs = [["Analog Out 1/2", 1],
               ["Analog Out 3/4", 2],
               ["Analog Out 5/6", 3],
               ["Analog Out 7/8", 4]]

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

    def getDisplayTitle(self):
        return 'FP10'

    def buildMixer(self):
        self.Volumes = {}
        self.Mutes = {}
        self.Pannings = {}

        layout = QGridLayout(self)

        for i in range(len(self.outputs)):
            self.addVolume(self.outputs[i][0], self.outputs[i][1], i, layout)

    def addVolume(self, name, bid, col, layout):
        label = QLabel(self)
        label.setText(name)
        l_dial = self.getDial()
        r_dial = self.getDial()
        l_sld = self.getSlider()
        r_sld = self.getSlider()
        l_mute = self.getMute()
        r_mute = self.getMute()
        link = self.getLink()

        layout.addWidget(label,  1, col * 2, 1, 2, Qt.AlignHCenter)
        layout.addWidget(l_dial, 2, col * 2,       Qt.AlignHCenter)
        layout.addWidget(r_dial, 2, col * 2 + 1,   Qt.AlignHCenter)
        layout.addWidget(l_sld,  3, col * 2,       Qt.AlignHCenter)
        layout.addWidget(r_sld,  3, col * 2 + 1,   Qt.AlignHCenter)
        layout.addWidget(l_mute, 4, col * 2,       Qt.AlignHCenter)
        layout.addWidget(r_mute, 4, col * 2 + 1,   Qt.AlignHCenter)
        layout.addWidget(link,   5, col * 2, 1, 2, Qt.AlignHCenter)
        
        path = "/Mixer/Feature_Volume_%d" % bid
        self.Volumes[l_sld] = [path, 1, r_sld, l_mute, link]
        self.Volumes[r_sld] = [path, 2, l_sld, r_mute, link]
        self.Mutes[l_mute] = [r_mute, l_sld]
        self.Mutes[r_mute] = [l_mute, r_sld]

        path = "/Mixer/Feature_LRBalance_%d" % bid
        self.Pannings[l_dial] = [path, 1]
        self.Pannings[r_dial] = [path, 2]
        
    # widget helper functions
    def getSlider(self):
        sld = QSlider(self)
        sld.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        sld.setMinimum(0)
        sld.setMaximum(99)
        sld.setPageStep(10)
        sld.setPageStep(10)
        sld.setMinimumHeight(50)
        sld.setTickInterval(25)
        sld.setTickPosition(QSlider.TicksBothSides)
        return sld

    def getDial(self):
        dial = QDial(self)
        dial.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        dial.setMinimumSize(50, 50)
        dial.setMaximumHeight(50)
        dial.setNotchTarget(25)
        dial.setNotchesVisible(True)
        dial.setMinimum(0)
        dial.setMaximum(99)
        dial.setPageStep(10)
        dial.setPageStep(10)
        return dial

    def getLink(self):
        link = QToolButton(self)
        link.setText("link")
        link.setCheckable(True)
        link.setMinimumWidth(100)
        link.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        return link

    def getMute(self):
        mute = QToolButton(self)
        mute.setText("mute")
        mute.setCheckable(True)
        mute.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        return mute

    def initValues(self):
        for ctl, params in self.Volumes.items():
           path = params[0]
           idx  = params[1]
           pair = params[2]
           mute = params[3]
           link = params[4]

           db = self.hw.getContignuous(path, idx)
           vol = self.db2vol(db)
           ctl.setValue(int(vol))
           ctl.valueChanged.connect(self.updateVolume)

           if vol == 0:
               mute.setChecked(True)

           if idx == 2:
               pair_db = self.hw.getContignuous(path, 1)
               if pair_db == db:
                   link.setChecked(True)

        #       Right - Center - Left
        # 0x8000 - 0x0000 - 0x0001 - 0x7FFE
        #        ..., -1, 0, +1, ...
        for ctl, params in self.Pannings.items():
            path = params[0]
            idx = params[1]

            val = self.hw.getContignuous(path, idx)
            state = -(val / 0x7FFE) * 50 + 50
            ctl.setValue(int(state))
            ctl.valueChanged.connect(self.updatePanning)

        for ctl, params in self.Mutes.items():
            ctl.clicked.connect(self.updateMute)

        return

    # helper functions
    def vol2db(self, vol):
        return (log10(vol + 1) - 2) * 16384

    def db2vol(self, db):
        return pow(10, db / 16384 + 2) - 1

    def updateVolume(self, vol):
        sender = self.sender()
        path   = self.Volumes[sender][0]
        idx    = self.Volumes[sender][1]
        pair   = self.Volumes[sender][2]
        mute   = self.Volumes[sender][3]
        link   = self.Volumes[sender][4]

        if mute.isChecked():
                return

        db = self.vol2db(vol)
        self.hw.setContignuous(path, db, idx)

        if link.isChecked():
           pair.setValue(vol)

    def updatePanning(self, state):
        sender = self.sender()
        path = self.Pannings[sender][0]
        idx = self.Pannings[sender][1]
        val  = (state - 50) * 0x7FFE / -50
        self.hw.setContignuous(path, idx, val)

    def updateMute(self, state):
        sender = self.sender()
        pair = self.Mutes[sender][0]
        sld = self.Mutes[sender][1]

        path = self.Volumes[sld][0]
        idx = self.Volumes[sld][1]
        pair_sld = self.Volumes[sld][2]
        link = self.Volumes[sld][4]

        if state:
            db = 0x8000
            vol = 0
        else:
            db = 0x0000
            vol = 99

        self.hw.setContignuous(path, db, idx)
        sld.setValue(vol)
        sld.setDisabled(state)

        if link.isChecked():
            if idx == 1:
               idx = 2
            else:
               idx = 1
            self.hw.setContignuous(path, db, idx)
            pair.setChecked(state)
            pair_sld.setValue(vol)
            pair_sld.setDisabled(state)




# Copyright (C) 2005-2008 by Pieter Palmers
#               2007-2009 by Arnold Krille
#
# This file is part of FFADO
# FFADO = Free FireWire (pro-)audio drivers for Linux
#
# FFADO is based upon FreeBoB.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import ctypes
import datetime
import os

from ffado.config import *

import subprocess

# from PyQt4.QtCore import QObject, QTimer, Qt
# from PyQt4.QtGui import *
from ffado.import_pyqt import *

from ffado.dbus_util import *

from ffado.panelmanager import PanelManager

from ffado.logginghandler import *

"""Just a small helper to ask the retry-question without a blocking messagebox"""
class StartDialog(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.setObjectName("Restart Dialog")
        self.label = QLabel("<qt>Somehow the connection to the dbus-service of FFADO couldn't be established.<p>\nShall we take another try?</qt>",self)
        self.button = QPushButton("Retry", self)
        self.layout = QGridLayout(self)
        self.layout.setContentsMargins( 50, 10, 50, 10 )
        self.layout.addWidget(self.label, 0, 0, Qt.AlignHCenter|Qt.AlignBottom)
        self.layout.addWidget(self.button, 1, 0, Qt.AlignHCenter|Qt.AlignTop)

class FFADOWindow(QMainWindow):
    def __init__(self, parent):
        QMainWindow.__init__(self, parent)

        self.textlogger = QTextLogger(self)
        dock = QDockWidget("Log Messages",self)
        dock.setWidget(self.textlogger.textedit)
        logging.getLogger('').addHandler(self.textlogger)
        self.addDockWidget(Qt.BottomDockWidgetArea, dock)

        self.statuslogger = QStatusLogger(self, self.statusBar(), 20)
        logging.getLogger('').addHandler(self.statuslogger)

        self.settings = QtCore.QSettings(self)

        self.manager = PanelManager(self)
        self.manager.connectionLost.connect(self.connectToDBUS)

        filemenu = self.menuBar().addMenu("&File")
        self.openaction = QAction(QIcon.fromTheme("document-open"),"&Open", self)
        self.openaction.setShortcut(self.tr("Ctrl+O"))
        self.openaction.setEnabled(False)
        self.openaction.triggered.connect(self.manager.readSettings)
        filemenu.addAction(self.openaction)
        self.saveaction = QAction(QIcon.fromTheme("document-save-as"),"&Save as...", self)
        self.saveaction.setShortcut(self.tr("Ctrl+S"))
        self.saveaction.setEnabled(False)
        self.saveaction.triggered.connect(self.manager.saveSettings)
        filemenu.addAction(self.saveaction)
        self.quitaction = QAction(QIcon.fromTheme("application-exit"),"&Quit", self)
        self.quitaction.setShortcut(self.tr("Ctrl+q"))
        self.quitaction.triggered.connect(self.close)
        filemenu.addAction(self.quitaction)

        self.editmenu = self.menuBar().addMenu("&View")

        self.thememenu = self.editmenu.addMenu("Theme")
        themes = QStyleFactory.keys()
        self.menuTheme = {}
        for theme in themes:
            self.menuTheme[theme] = QAction(QIcon.fromTheme("preferences-desktop-theme"), theme, self )
            self.menuTheme[theme].setCheckable(True)

            if (ffado_python3 and (self.style().objectName().lower() == theme.lower()) or
                    not(ffado_python3) and (self.style().objectName().toLower() == theme.toLower() if ffado_pyqt_version == 4 else
                                            self.style().objectName().lower() == theme.lower())):
                self.menuTheme[theme].setDisabled(True)
                self.menuTheme[theme].setChecked(True)
            self.menuTheme[theme].triggered.connect(self.switchTheme )
            self.thememenu.addAction( self.menuTheme[theme] )

        conftheme = self.settings.value("window/theme", "ukui-dark")
        contains = False
        for theme in self.menuTheme:
            if theme.__str__() == conftheme:
                contains = True
                break
        if contains:
            for theme in self.menuTheme:
                if theme.__str__() != conftheme:
                    self.menuTheme[theme].setChecked(False)
                    self.menuTheme[theme].setDisabled(False)
            for theme in self.menuTheme:
                if theme.__str__() == conftheme:
                    self.menuTheme[theme].setDisabled(True)
                    QApplication.setStyle(QStyleFactory.create(theme))
                    self.settings.setValue("window/theme", theme.__str__())

        self.updateaction = QAction(QIcon.fromTheme("view-refresh"),"&Update Mixer Panels", self)
        self.updateaction.setEnabled(False)
        self.updateaction.triggered.connect(self.manager.updatePanels)
        self.editmenu.addAction(self.updateaction)
        self.refreshaction = QAction(QIcon.fromTheme("view-refresh"),"&Refresh Current Panels", self)
        self.refreshaction.triggered.connect(self.manager.refreshPanels)
        self.editmenu.addAction(self.refreshaction)
        self.editmenu.addSeparator()
        self.devices = {}

        helpmenu = self.menuBar().addMenu( "&Help" )
        self.aboutaction = QAction(QIcon.fromTheme("help-about"), "About &FFADO", self )
        self.aboutaction.triggered.connect(self.aboutFFADO)
        helpmenu.addAction( self.aboutaction )
        self.aboutqtaction = QAction(QIcon.fromTheme("help-about"),  "About &Qt", self )
        self.aboutqtaction.triggered.connect(QApplication.instance().aboutQt)
        helpmenu.addAction( self.aboutqtaction )

        log.info( "Starting up" )

        QTimer.singleShot( 1, self.tryStartDBUSServer )

    def __del__(self):
        log.info("__del__")
        del self.manager
        log.info("__del__ finished")

    def switchTheme(self, checked) :
        for theme in self.menuTheme :
            if not self.menuTheme[theme].isEnabled() :
                self.menuTheme[theme].setChecked(False)
                self.menuTheme[theme].setDisabled(False)
        for theme in self.menuTheme :
            if self.menuTheme[theme].isChecked() :
                self.menuTheme[theme].setDisabled(True)
                QApplication.setStyle(QStyleFactory.create(theme))
                self.settings.setValue("window/theme", theme.__str__())

    def closeEvent(self, event):
        log.info("closeEvent()")
        event.accept()

    def connectToDBUS(self):
        log.info("connectToDBUS")
        try:
            self.setupDeviceManager()
        except dbus.DBusException as ex:
            log.error("Could not communicate with the FFADO DBus service...")
            if not hasattr(self,"retry"):
                self.retry = StartDialog(self)
                self.retry.button.clicked.connect(self.tryStartDBUSServer)
            if hasattr(self, "retry"):
                self.manager.setParent(None)
                self.setCentralWidget(self.retry)
                self.retry.setEnabled(True)

    def tryStartDBUSServer(self):
        try:
            self.setupDeviceManager()
        except dbus.DBusException as ex:
            if hasattr(self, "retry"):
                self.retry.setEnabled(False)
            subprocess.Popen(['ffado-dbus-server', '-v3'], close_fds=True).pid
            QTimer.singleShot(5000, self.connectToDBUS)

    def setupDeviceManager(self):
        devmgr = DeviceManagerInterface(FFADO_DBUS_SERVER, FFADO_DBUS_BASEPATH)
        self.manager.setManager(devmgr)
        if hasattr(self, "retry"):
            self.retry.setParent(None)
        self.setCentralWidget(self.manager)
        self.updateaction.setEnabled(True)

    def aboutFFADO(self):
        QMessageBox.about( self, "About FFADO", """
<h1><a href="http://ffado.org">ffado.org</a></h1>

<p>{ffado_version}</p>

<p><a href="http://ffado.org">FFADO</a> is the new approach to have FireWire audio on Linux.</p>

        <p>&copy; 2006-2021 by the FFADO developers<br />ffado is licensed under the GPLv3, for the full license text see <a href="http://www.gnu.org/licenses/">www.gnu.org/licenses</a> or the LICENSE.* files shipped with ffado.</p>

<p>FFADO developers are:<ul>
<li>Pieter Palmers
<li>Daniel Wagner
<li>Jonathan Woithe
<li>Arnold Krille
<li>Philippe Carriere
<li>Takashi Sakamoto
</ul>
with contributions from:<ul>
<li>Adrian Knoth
<li>Stefan Richter
<li>Jano Svitok
<li>Pander Musubi
</ul>
        """.format(ffado_version=get_ffado_version(), thisyear=datetime.datetime.now().year))

def get_ffado_version():
    try:
        # call the C function ffado_get_version() to figure out the version
        lib = ctypes.cdll.LoadLibrary('libffado.so')
        func = ctypes.CFUNCTYPE(ctypes.c_char_p)
        ffado_get_version = func(('ffado_get_version', lib))
        return ffado_get_version()
    except:
        return "libffado"

def get_lock(process_name):
    import socket
    import sys

    global lock_socket
    lock_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        lock_socket.bind('\0' + process_name)
        # Lock acquired
    except socket.error:
        print( 'ffado-mixer instance is already running' )
        sys.exit()


def ffadomain(args):
    #set up logging
    import logging
    logging.basicConfig( datefmt="%H:%M:%S", format="%(asctime)s %(name)-16s %(levelname)-8s %(message)s" )

    if DEBUG:
        debug_level = logging.DEBUG
    else:
        debug_level = logging.INFO

    get_lock('ffado-mixer')

    # Very simple command line option parser
    if (len(args) > 1) and (args[1] == "-b" or args[1] == "--bypassdbus"):
        ffado.config.bypassdbus = True

    # main loggers:
    logging.getLogger('main').setLevel(debug_level)
    logging.getLogger('dbus').setLevel(debug_level)
    logging.getLogger('registration').setLevel(debug_level)
    logging.getLogger('panelmanager').setLevel(debug_level)
    logging.getLogger('configparser').setLevel(logging.INFO)

    # widgets:
    logging.getLogger('matrixmixer').setLevel(debug_level)
    logging.getLogger('crossbarrouter').setLevel(debug_level)

    # mixers:
    logging.getLogger('audiofire').setLevel(debug_level)
    logging.getLogger('bridgeco').setLevel(debug_level)
    logging.getLogger('edirolfa101').setLevel(debug_level)
    logging.getLogger('edirolfa66').setLevel(debug_level)
    logging.getLogger('motu').setLevel(debug_level)
    logging.getLogger('rme').setLevel(debug_level)
    logging.getLogger('phase24').setLevel(debug_level)
    logging.getLogger('phase88').setLevel(debug_level)
    logging.getLogger('quatafire').setLevel(debug_level)
    logging.getLogger('saffirebase').setLevel(debug_level)
    logging.getLogger('saffire').setLevel(debug_level)
    logging.getLogger('saffirepro').setLevel(debug_level)

    logging.getLogger('global').setLevel(debug_level)

    log = logging.getLogger('main')
    log.debug("Using %s with Qt: %s PyQt: %s" % (get_ffado_version(), QtCore.QT_VERSION_STR, QtCore.PYQT_VERSION_STR))

    app = QApplication(args)
    app.setWindowIcon( QIcon( SHAREDIR + "/icons/hi64-apps-ffado.png" ) )

    app.setOrganizationName("FFADO")
    app.setOrganizationDomain("ffado.org")
    app.setApplicationName("ffado-mixer")

    mainwindow = FFADOWindow(None)


    # rock & roll
    mainwindow.show()
    return app.exec_()

if __name__ == "__main__":
    import sys
    sys.exit(ffadomain(sys.argv))

#
# vim: ts=4 sw=4 et



#
# Copyright (C) 2005-2008 by Pieter Palmers
#
# This file is part of FFADO
# FFADO = Free FireWire (pro-)audio drivers for Linux
#
# FFADO is based upon FreeBoB.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# from PyQt4.QtGui import QWidget
from ffado.import_pyqt import *

from ffado.config import *

import logging
log = logging.getLogger('phase24')

class Phase24Control(QWidget):
    def __init__(self,parent = None):
        QWidget.__init__(self,parent)
        uicLoad("ffado/mixer/phase24", self)

        self.VolumeControls={
            'analogin' :      ['/Mixer/Feature_Volume_6', self.sldLineIn],
            'spdifin' :       ['/Mixer/Feature_Volume_7', self.sldSPDIFIn],
            'out12' :         ['/Mixer/Feature_Volume_3', self.sldInput12],
            'out34' :         ['/Mixer/Feature_Volume_4', self.sldInput34],
            'outspdif' :      ['/Mixer/Feature_Volume_5', self.sldSPDIFOut],
            }

        self.SelectorControls={
            'outsource12':    ['/Mixer/Selector_1', self.cmbOutSource12], 
            'outsource34':    ['/Mixer/Selector_2', self.cmbOutSource34], 
            'outsourcespdif': ['/Mixer/Selector_3', self.cmbOutSourceSPDIF], 
            'syncsource':     ['/Mixer/Selector_4', self.cmbSetSyncSource], 
        }

    # public slot
    def setVolume12(self,a0):
        self.setVolume('out12', a0)

    # public slot
    def setVolume34(self,a0):
        self.setVolume('out34', a0)

    # public slot
    def setVolumeLineIn(self,a0):
        self.setVolume('analogin', a0)

    # public slot
    def setVolumeSPDIFOut(self,a0):
        self.setVolume('outspdif', a0)

    # public slot
    def setVolumeSPDIFIn(self,a0):
        self.setVolume('spdifin', a0)

    # public slot
    def setVolumeMaster(self,a0):
        if self.isPhaseX24:
            return
        self.setVolume('master', a0)

    # public slot
    def setLineLevel(self,a0):
        log.debug("setting line level to %d" % (a0 * -768))
        self.hw.setContignuous('/Mixer/Feature_Volume_2', a0 * -768)

    # public slot
    def setFrontLevel(self,a0):
        if self.isPhaseX24:
            return
        if(a0 == 0):
            log.debug("setting front level to %d" % (0))
            self.hw.setContignuous('/Mixer/Feature_Volume_8', 0)
        else:
            log.debug("setting front level to %d" % (1536))
            self.hw.setContignuous('/Mixer/Feature_Volume_8', 1536)

    # public slot
    def setOutSource12(self,a0):
        self.setSelector('outsource12', a0)

    # public slot
    def setOutSource34(self,a0):
        self.setSelector('outsource34', a0)

    # public slot
    def setOutSourceSPDIF(self,a0):
        self.setSelector('outsourcespdif', a0)

    # public slot
    def setSyncSource(self,a0):
        self.setSelector('syncsource', a0)

    def setVolume(self,a0,a1):
            name=a0
            vol = -a1
            log.debug("setting %s volume to %d" % (name, vol))
            self.hw.setContignuous(self.VolumeControls[name][0], vol)

    def setSelector(self,a0,a1):
            name=a0
            state = a1
            log.debug("setting %s state to %d" % (name, state))
            self.hw.setDiscrete(self.SelectorControls[name][0], state)

    def initValues(self):
            self.modelId = self.configrom.getModelId()
            if self.modelId == 0x00000007:
                self.isPhaseX24 = True
            else:
                self.isPhaseX24 = False

            if self.isPhaseX24:
                self.setWindowTitle("Terratec Phase X24 Control")
                self.cmbFrontLevel.setEnabled(False)
                self.sldMaster.setEnabled(False)
            else:
                self.setWindowTitle("Terratec Phase 24 Control")

                self.VolumeControls['master'] = ['/Mixer/Feature_Volume_1', self.sldMaster]
                self.sldMaster.setEnabled(True)

                self.cmbFrontLevel.setEnabled(True)
                val = self.hw.getContignuous('/Mixer/Feature_Volume_8')
                if val > 0:
                    self.cmbFrontLevel.setCurrentIndex(1)
                else:
                    self.cmbFrontLevel.setCurrentIndex(0)

            for name, ctrl in self.VolumeControls.items():
                vol = self.hw.getContignuous(ctrl[0])
                log.debug("%s volume is %d" % (name , vol))
                ctrl[1].setValue(-vol)

            for name, ctrl in self.SelectorControls.items():
                state = self.hw.getDiscrete(ctrl[0])
                log.debug("%s state is %d" % (name , state))
                ctrl[1].setCurrentIndex(state)

            val = self.hw.getContignuous('/Mixer/Feature_Volume_2')/-768
            if val > 4:
                self.cmbLineLevel.setCurrentIndex(4)
            else:
                self.cmbLineLevel.setCurrentIndex(val)

# vim: et
